# -*- coding: utf-8 -*-

import abc
import glob
import os
import shutil
import sys
import traceback

from macdaily.cls.command import Command
from macdaily.util.compat import pathlib, subprocess
from macdaily.util.const.term import bold, flash, purple_bg, red, red_bg, reset, under, yellow
from macdaily.util.tools.make import make_stderr
from macdaily.util.tools.misc import date
from macdaily.util.tools.print import print_info, print_scpt, print_term, print_text
from macdaily.util.tools.script import run


class CaskCommand(Command):

    @property
    def mode(self):
        return 'cask'

    @property
    def name(self):
        return 'Homebrew Casks'

    @property
    def desc(self):
        return ('Caskroom binary', 'Caskroom binaries')

    def _check_exec(self):
        try:
            subprocess.check_call(['brew', 'command', 'cask'],
                                  stdout=subprocess.DEVNULL, stderr=make_stderr(self._vflag))
        except subprocess.CalledProcessError:
            print_text(traceback.format_exc(), self._file, redirect=self._vflag)
            print('macdaily-{}: {}{}cask{}: command not found'.format(self.cmd, red_bg, flash, reset), file=sys.stderr)
            text = ('macdaily-{}: {}cask{}: you may find Caskroom on '
                    '{}{}https://caskroom.github.io{}, '
                    'or install Caskroom through following command -- '
                    "`{}brew tap homebrew/cask{}'".format(self.cmd, red, reset, purple_bg, under, reset, bold, reset))
            print_term(text, self._file, redirect=self._qflag)
            return False
        self._var__exec_path = shutil.which('brew')
        return True

    @abc.abstractmethod
    def _parse_args(self, namespace):
        super()._parse_args(namespace)
        self._force = namespace.get('force', False)  # pylint: disable=attribute-defined-outside-init
        self._merge = namespace.get('merge', False)  # pylint: disable=attribute-defined-outside-init
        self._no_cleanup = namespace.get('no_cleanup', False)  # pylint: disable=attribute-defined-outside-init
        self._verbose = namespace.get('verbose', False)  # pylint: disable=attribute-defined-outside-init

    def _loc_exec(self):
        self._exec = {self._var__exec_path}
        del self._var__exec_path

    def _check_pkgs(self, path):
        text = 'Listing installed {}'.format(self.desc[1])
        print_info(text, self._file, redirect=self._vflag)

        argv = [path, 'cask', 'list']
        args = ' '.join(argv)
        print_scpt(args, self._file, redirect=self._vflag)
        with open(self._file, 'a') as file:
            file.write('Script started on {}\n'.format(date()))
            file.write('command: {!r}\n'.format(args))

        try:
            proc = subprocess.check_output(argv, stderr=make_stderr(self._vflag))
        except subprocess.CalledProcessError:
            print_text(traceback.format_exc(), self._file, redirect=self._vflag)
            _real_pkgs = set()
        else:
            context = proc.decode()
            _real_pkgs = set(context.split())
            print_text(context, self._file, redirect=self._vflag)
        finally:
            with open(self._file, 'a') as file:
                file.write('Script done on {}\n'.format(date()))

        text = 'Checking existence of specified packages'
        print_info(text, self._file, redirect=self._vflag)

        _temp_pkgs = list()
        _lost_pkgs = list()
        for package in self._packages:
            if package in _real_pkgs:
                _temp_pkgs.append(package)
            else:
                _lost_pkgs.append(package)
        self._lost.extend(_lost_pkgs)

        self._var__real_pkgs = set(_real_pkgs)  # pylint: disable=attribute-defined-outside-init
        self._var__lost_pkgs = set(_lost_pkgs)  # pylint: disable=attribute-defined-outside-init
        self._var__temp_pkgs = set(_temp_pkgs)  # pylint: disable=attribute-defined-outside-init

    def _proc_renew(self, path):
        text = 'Updating Homebrew database'
        print_info(text, self._file, redirect=self._qflag)

        argv = [path, 'update']
        if self._force:
            argv.append('--force')
        if self._merge:
            argv.append('--merge')
        if self._quiet:
            argv.append('--quiet')
        if self._verbose:
            argv.append('--verbose')
        print_scpt(' '.join(argv), self._file, redirect=self._qflag)
        run(argv, self._file, redirect=self._qflag)

    def _proc_cleanup(self):
        if self._no_cleanup:
            return

        text = 'Pruning caches and archives'
        print_info(text, self._file, redirect=self._qflag)

        if not os.path.isdir(self._disk_dir):
            text = ('macdaily-{}: {}cask{}: '
                    'archive directory {}{}{} not found'.format(self.cmd, yellow, reset, bold, self._disk_dir, reset))
            print_term(text, self._file, redirect=self._vflag)
            return

        argv = ['brew', 'cask', 'cleanup']
        if self._verbose:
            argv.append('--verbose')
        if self._quiet:
            argv.append('--quiet')
        print_scpt(' '.join(argv), self._file, redirect=self._qflag)

        path_cask = os.path.join(self._disk_dir, 'Homebrew', 'Cask')
        path_down = os.path.join(self._disk_dir, 'Homebrew', 'download')

        pathlib.Path(path_cask).mkdir(parents=True, exist_ok=True)
        pathlib.Path(path_down).mkdir(parents=True, exist_ok=True)

        for path in self._exec:
            argv = [path, '--cache']
            args = ' '.join(argv)
            print_scpt(args, self._file, redirect=self._vflag)
            with open(self._file, 'a') as file:
                file.write('Script started on {}\n'.format(date()))
                file.write('command: {!r}\n'.format(args))

            fail = False
            try:
                proc = subprocess.check_output(argv, stderr=make_stderr(self._vflag))
            except subprocess.CalledProcessError:
                print_text(traceback.format_exc(), self._file, redirect=self._vflag)
                fail = True
            else:
                context = proc.decode()
                print_text(context, self._file, redirect=self._vflag)
            finally:
                with open(self._file, 'a') as file:
                    file.write('Script done on {}\n'.format(date()))
            if fail:
                continue

            cache = context.strip()
            if os.path.isdir(cache):
                argv = [path, 'cask', 'caches', 'archive']
                if self._verbose:
                    argv.append('--verbose')
                if self._quiet:
                    argv.append('--quiet')
                print_scpt(' '.join(argv), self._file, redirect=self._qflag)

                file_list = list()
                link_list = glob.glob(os.path.join(cache, 'Cask/*'))
                cask_list = [os.path.realpath(name) for name in link_list]
                for link in link_list:
                    file_list.append(link)
                    try:
                        shutil.move(link, path_cask)
                    except (shutil.Error, FileExistsError):
                        os.remove(link)
                for cask in filter(lambda p: os.path.splitext(p)[1] != '.incomplete', cask_list):
                    try:
                        shutil.move(cask, path_down)
                    except (shutil.Error, FileExistsError):
                        os.remove(cask)
                    file_list.append(cask)
                print_text(os.linesep.join(sorted(file_list)), self._file, redirect=self._vflag)



# -*- coding: utf-8 -*-

import argparse
import sys

from macdaily.util.const.macro import STR_HELP
from macdaily.util.const.macro import VERSION as __version__


def get_help_parser():
    #######################################################
    # Help CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    parser = argparse.ArgumentParser(prog='macdaily-help',
                                     description='MacDaily Usage Information Manual',
                                     usage='macdaily help [options] <cmd-selection> ...',
                                     epilog=STR_HELP)
    parser.add_argument('-V', '--version', action='version', version=__version__)

    spec_group = parser.add_argument_group(title='specification arguments')
    spec_group.add_argument('command', action='store', nargs='?', metavar='CMD',
                            help=("display manual information about such command".format()))

    return parser


def parse_args(argv=None):
    if argv is None:
        argv = sys.argv[1:] or ['--help']

    # main parser process
    main_parser = get_help_parser()
    main_args = main_parser.parse_args(argv)

    return main_args



# -*- coding: utf-8 -*-

import argparse
import sys

from macdaily.util.const.macro import STR_POSTINSTALL
from macdaily.util.const.macro import VERSION as __version__
from macdaily.util.const.term import bold, reset


def get_postinstall_parser():
    #######################################################
    # Postinstall CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    parser = argparse.ArgumentParser(prog='macdaily-postinstall',
                                     description='Homebrew Cask Postinstall Automator',
                                     usage='macdaily postinstall [options] ...',
                                     epilog=STR_POSTINSTALL)
    parser.add_argument('-V', '--version', action='version', version=__version__)

    spec_group = parser.add_argument_group(title='specification arguments')
    spec_group.add_argument('-s', '--startswith', action='store', metavar='PREFIX',
                            help='postinstall procedure starts from such formula, sort in initial alphabets')
    spec_group.add_argument('-e', '--endswith', action='store', metavar='SUFFIX',
                            help='postinstall procedure ends after such formula, sort in initial alphabets')
    spec_group.add_argument('-p', '--packages', action='append', nargs='+', default=list(), metavar='FORM',
                            help='name of Homebrew formulae to postinstall')

    genl_group = parser.add_argument_group(title='general arguments')
    genl_group.add_argument('-a', '--all', action='store_true',
                            help='postinstall all Homebrew formulae installed through Homebrew')
    genl_group.add_argument('-q', '--quiet', action='store_true',
                            help='run in quiet mode, with no output information')
    genl_group.add_argument('-v', '--verbose', action='store_true',
                            help='run in verbose mode, with detailed output information')
    genl_group.add_argument('-l', '--show-log', action='store_true',
                            help='open log in Console.app upon completion of command')
    genl_group.add_argument('-y', '--yes', action='store_true',
                            help='yes for all selections')
    genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                            help='do not run cleanup process')

    misc_group = parser.add_argument_group(title='miscellaneous arguments')
    misc_group.add_argument('-L', '--logging', action='store', default=str(), metavar='ARG',
                            help="options for `{}brew list{}' command".format(bold, reset))
    misc_group.add_argument('-P', '--postinstall', action='store', default=str(), metavar='ARG',
                            help="options for `{}brew postinstall <formula>{}' command".format(bold, reset))

    return parser


def parse_args(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    # parser process
    parser = get_postinstall_parser()
    args = parser.parse_args(argv or ['--help'])

    return args



# -*- coding: utf-8 -*-

import argparse
import re
import sys

from macdaily.util.const.macro import CMD_REINSTALL, MAP_DICT, STR_BREW, STR_CASK, STR_REINSTALL
from macdaily.util.const.macro import VERSION as __version__
from macdaily.util.const.term import bold, reset


def get_reinstall_parser():
    #######################################################
    # Reinstall CLI
    #   * options
    #       - optional arguments
    #       - general arguments
    #       - package arguments
    #       - control arguments
    #   * specifications
    #######################################################

    parser = argparse.ArgumentParser(prog='macdaily-reinstall',
                                     description='Automated macOS Package Reinstaller',
                                     usage='macdaily reinstall [options] <mode-selection> ...',
                                     epilog=STR_REINSTALL)
    parser.add_argument('-V', '--version', action='version', version=__version__)
    parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    genl_group = parser.add_argument_group(title='general arguments')
    genl_group.add_argument('-a', '--all', action='store_true',
                            help='reinstall all packages installed through Homebrew, Caskroom, and etc')
    genl_group.add_argument('-q', '--quiet', action='store_true',
                            help='run in quiet mode, with no output information')
    genl_group.add_argument('-v', '--verbose', action='store_true',
                            help='run in verbose mode, with detailed output information')
    genl_group.add_argument('-l', '--show-log', action='store_true',
                            help='open log in Console.app upon completion of command')
    genl_group.add_argument('-y', '--yes', action='store_true',
                            help='yes for all selections')
    genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                            help='do not run cleanup process')

    pkgs_group = parser.add_argument_group(title='package arguments',
                                           description='options used to specify packages of each mode')
    pkgs_group.add_argument('--brew', action='append', nargs='+', default=list(), metavar='FORM', dest='brew_pkgs',
                            help='name of Homebrew formulae to reinstall')
    pkgs_group.add_argument('--cask', action='append', nargs='+', default=list(), metavar='CASK', dest='cask_pkgs',
                            help='name of Caskroom binaries to reinstall')

    ctrl_group = parser.add_argument_group(title='control arguments',
                                           description='options used to disable reinstall of certain mode')
    ctrl_group.add_argument('--no-brew', action='store_true',
                            help='do not reinstall Homebrew formulae')
    ctrl_group.add_argument('--no-cask', action='store_true',
                            help='do not reinstall Caskroom binaries')

    parser.add_argument_group(title='mode selection',
                              description=('reinstall existing packages installed through a specified method, '
                                           'e.g.: brew, cask'))

    return parser


def get_brew_parser():
    #######################################################
    # Brew Reinstall CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    brew_parser = argparse.ArgumentParser(prog='macdaily-reinstall-brew',
                                          description='Automated Homebrew Formula Reinstaller',
                                          usage='macdaily reinstall brew [options] <formulae> ...',
                                          epilog=STR_BREW)
    brew_parser.add_argument('-V', '--version', action='version', version=__version__)
    brew_parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    brew_spec_group = brew_parser.add_argument_group(title='specification arguments')
    brew_spec_group.add_argument('-s', '--startswith', action='store', metavar='PREFIX',
                                 help='reinstall procedure starts from such formula, sort in initial alphabets')
    brew_spec_group.add_argument('-e', '--endswith', action='store', metavar='SUFFIX',
                                 help='reinstall procedure ends after such formula, sort in initial alphabets')
    brew_spec_group.add_argument('-p', '--packages', action='append', nargs='+', default=list(), metavar='FORM',
                                 help='name of Homebrew formulae to reinstall')

    brew_genl_group = brew_parser.add_argument_group(title='general arguments')
    brew_genl_group.add_argument('-a', '--all', action='store_true',
                                 help='reinstall all Homebrew formulae installed through Homebrew')
    brew_genl_group.add_argument('-q', '--quiet', action='store_true',
                                 help='run in quiet mode, with no output information')
    brew_genl_group.add_argument('-v', '--verbose', action='store_true',
                                 help='run in verbose mode, with detailed output information')
    brew_genl_group.add_argument('-y', '--yes', action='store_true',
                                 help='yes for all selections')
    brew_genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                                 help='do not run cleanup process')

    brew_misc_group = brew_parser.add_argument_group(title='miscellaneous arguments')
    brew_misc_group.add_argument('-L', '--logging', action='store', default=str(), metavar='ARG',
                                 help="options for `{}brew list{}' command".format(bold, reset))
    brew_misc_group.add_argument('-R', '--reinstall', action='store', default=str(), metavar='ARG',
                                 help="options for `{}brew reinstall <formula>{}' command".format(bold, reset))

    return brew_parser


def get_cask_parser():
    #######################################################
    # Cask Reinstall CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    cask_parser = argparse.ArgumentParser(prog='macdaily-reinstall-cask',
                                          description='Automated Homebrew Cask Reinstaller',
                                          usage='macdaily reinstall cask [options] <casks> ...',
                                          epilog=STR_CASK)
    cask_parser.add_argument('-V', '--version', action='version', version=__version__)
    cask_parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    cask_spec_group = cask_parser.add_argument_group(title='specification arguments')
    cask_spec_group.add_argument('-s', '--startswith', action='store', metavar='PREFIX',
                                 help='reinstall procedure starts from such binary, sort in initial alphabets')
    cask_spec_group.add_argument('-e', '--endswith', action='store', metavar='SUFFIX',
                                 help='reinstall procedure ends after such binary, sort in initial alphabets')
    cask_spec_group.add_argument('-f', '--force', action='store_true',
                                 help='reinstall even if the Cask does not appear to be present')
    cask_spec_group.add_argument('-t', '--no-quarantine', action='store_true',
                                 help='prevent Gatekeeper from enforcing its security restrictions on the Cask')
    cask_spec_group.add_argument('-p', '--packages', action='append', nargs='+', default=list(), metavar='CASK',
                                 help='name of Caskroom binaries to reinstall')

    cask_genl_group = cask_parser.add_argument_group(title='general arguments')
    cask_genl_group.add_argument('-a', '--all', action='store_true',
                                 help='reinstall all Caskroom binaries installed through Homebrew')
    cask_genl_group.add_argument('-q', '--quiet', action='store_true',
                                 help='run in quiet mode, with no output information')
    cask_genl_group.add_argument('-v', '--verbose', action='store_true',
                                 help='run in verbose mode, with detailed output information')
    cask_genl_group.add_argument('-y', '--yes', action='store_true',
                                 help='yes for all selections')
    cask_genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                                 help='do not run cleanup process')

    cask_misc_group = cask_parser.add_argument_group(title='miscellaneous arguments')
    cask_misc_group.add_argument('-L', '--logging', action='store', default=str(), metavar='ARG',
                                 help="options for `{}brew cask list{}' command".format(bold, reset))
    cask_misc_group.add_argument('-R', '--reinstall', action='store', default=str(), metavar='ARG',
                                 help="options for `{}brew cask reinstall <cask>{}' command".format(bold, reset))

    return cask_parser


def parse_args(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    # main parser process
    main_parser = get_reinstall_parser()
    main_args = main_parser.parse_args(argv or ['--help'])

    def _update(d, e):
        for k, v in e.items():
            if k in d:
                if isinstance(v, list):
                    d[k].extend(v)
                else:
                    d[k] = v
            else:
                d[k] = v
        return d

    # recursively parse mode options
    more_opts = main_args.more_opts
    while more_opts:
        # traverse all extra options
        for index, option in enumerate(more_opts, start=1):
            # ignore ``--`` (end of option list)
            if option == '--':
                continue

            # check if legal mode
            parser_name = MAP_DICT.get(option.lower(), 'null')
            get_parser = globals().get('get_{}_parser'.format(parser_name))
            if get_parser is None:
                pattern = r'.*{}.*'.format(option)
                matches = "', '".format().join(filter(lambda s: re.match(pattern, s, re.IGNORECASE), CMD_REINSTALL))  # pylint: disable=cell-var-from-loop
                if matches:
                    main_parser.error('unrecognized arguments: {!r} '
                                      "(did you mean: '{}')".format(option, matches))
                else:
                    main_parser.error('unrecognized arguments: {!r}'.format(option))

            # parse mode arguments
            parser = get_parser()
            args = parser.parse_args(more_opts[index:] or ['--help'])

            # store/update parsed arguments
            opt_dict = getattr(main_args, option, dict())
            opt_dict.pop('more_opts', None)
            setattr(main_args, option, _update(opt_dict, vars(args)))

            # check for extra options
            more_opts = args.more_opts
            break

    delattr(main_args, 'more_opts')
    return main_args



# -*- coding: utf-8 -*-

import abc

from macdaily.cls.command import Command
from macdaily.util.tools.print import print_info


class UninstallCommand(Command):

    @property
    def cmd(self):
        return 'uninstall'

    @property
    def act(self):
        return ('uninstall', 'uninstalled', 'removed')

    @property
    def job(self):
        return ('uninstallation', 'uninstallation')

    def _run_proc(self):
        self._pkgs = list()
        self._fail = list()
        self._lost = list()
        for path in self._exec:
            text = 'Using {} executable {!r}'.format(self.name, path)
            print_info(text, self._file, redirect=self._qflag)
            if self._proc_logging(path):
                self._proc_uninstall(path)
            else:
                text = 'No {} to uninstall for executable {!r}'.format(self.desc[1], path)
                print_info(text, self._file, redirect=self._qflag)
            self._proc_fixmissing(path)
        self._proc_cleanup()

    def _proc_logging(self, path):
        if self._packages:
            self._check_pkgs(path)
            self._did_you_mean()
        else:
            self._check_list(path)
        return self._check_confirm(path)

    @abc.abstractmethod
    def _check_pkgs(self, path):
        self._var__temp_pkgs = self._packages  # pylint: disable=attribute-defined-outside-init

    @abc.abstractmethod
    def _check_list(self, path):
        self._var__temp_pkgs = set()  # pylint: disable=attribute-defined-outside-init

    @abc.abstractmethod
    def _proc_uninstall(self, path):
        pass



# -*- coding: utf-8 -*-

import functools
import os
import platform
import queue
import signal
import sys

from macdaily.util.compat import multiprocessing, subprocess, threading
from macdaily.util.const.macro import PASS, USER, USR
from macdaily.util.const.term import bold, dim, red, reset, under, yellow
from macdaily.util.error import ChildExit, IncorrectPassword, TimeExpired, UnsupportedOS
from macdaily.util.tools.make import make_pipe
from macdaily.util.tools.misc import kill
from macdaily.util.tools.print import print_term

# error-not-raised flag
ERR_FLAG = True
# func-not-called flag
FUNC_FLAG = True


def beholder(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        from macdaily.util.tools.get import get_logfile

        if platform.system() != 'Darwin':
            print_term('macdaily: error: script runs only on macOS', get_logfile())
            raise UnsupportedOS

        def _finale(epilogue):
            global FUNC_FLAG
            if FUNC_FLAG:
                FUNC_FLAG = False
                sys.stdout.write(reset)
                sys.stderr.write(reset)
                kill(os.getpid(), signal.SIGSTOP)
            return epilogue

        def _funeral(last_words):
            global ERR_FLAG
            ERR_FLAG = False
            # sys.tracebacklimit = 0
            sys.stdout.write(reset)
            sys.stderr.write(reset)
            kill(os.getpid(), signal.SIGKILL)
            print(last_words, file=sys.stderr)

        try:
            return _finale(func(*args, **kwargs))
        except KeyboardInterrupt:
            if ERR_FLAG:
                _funeral('macdaily: {}error{}: operation interrupted'.format(red, reset))
            raise
        except Exception:
            if ERR_FLAG:
                _funeral('macdaily: {}error{}: operation failed'.format(red, reset))
            raise
    return wrapper


def check(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        from macdaily.util.tools.get import get_logfile, get_boolean

        if get_boolean('NULL_PASSWORD'):
            return PASS

        SUDO_PASSWORD = os.getenv('SUDO_PASSWORD')
        if SUDO_PASSWORD is None:
            password = func(*args, **kwargs)
        else:
            password = SUDO_PASSWORD

        if not get_boolean('MACDAILY_NO_CHECK'):
            try:
                subprocess.run(['sudo', '--reset-timestamp'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                with make_pipe(password, redirect=True) as pipe:
                    subprocess.check_call(['sudo', '--stdin', '--prompt=""', "true"],
                                          stdin=pipe.stdout, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            except subprocess.CalledProcessError:
                global ERR_FLAG
                ERR_FLAG = False
                print_term('macdaily: {}error{}: incorrect password {}{!r}{} for '
                           '{}{}{} ({}{}{})'.format(red, reset, dim, password, reset, bold, USER, reset, under, USR, reset), get_logfile())
                raise IncorrectPassword from None
        return password
    return wrapper


def retry(default=None):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            from macdaily.util.tools.get import get_logfile

            if sys.stdin.isatty():  # pylint: disable=no-else-return
                return func(*args, **kwargs)
            else:
                # timeout interval
                from macdaily.util.tools.get import get_int
                TIMEOUT = get_int('TIMEOUT', 60)

                QUEUE = multiprocessing.Queue(1)
                kwargs['queue'] = QUEUE
                for _ in range(3):
                    proc = multiprocessing.Process(target=func, args=args, kwargs=kwargs)
                    timer = threading.Timer(TIMEOUT, function=proc.kill)
                    timer.start()
                    proc.start()
                    proc.join()
                    timer.cancel()
                    if proc.exitcode == 0:
                        break
                    if proc.exitcode != 9:
                        print_term('macdaily: {}error{}: function {!r} '
                                   'exits with exit status {} on child process'.format(yellow, reset, func.__qualname__, proc.exitcode), get_logfile())
                        raise ChildExit
                else:
                    print_term('macdaily: {}error{}: function {!r} '
                               'retry timeout after {} seconds'.format(red, reset, func.__qualname__, TIMEOUT), get_logfile())
                    raise TimeExpired
                try:
                    return QUEUE.get(block=False)
                except queue.Empty:
                    return default
        return wrapper
    return decorator



# -*- coding: utf-8 -*-

import abc
import shutil
import sys

from macdaily.cls.command import Command
from macdaily.util.const.term import bold, flash, red, red_bg, reset
from macdaily.util.tools.print import print_info, print_term


class SystemCommand(Command):

    @property
    def mode(self):
        return 'system'

    @property
    def name(self):
        return 'macOS Software Update'

    @property
    def desc(self):
        return ('system software', 'system software')

    def _check_exec(self):
        self._var__exec_path = shutil.which('softwareupdate')
        flag = (self._var__exec_path is not None)
        if not flag:
            print('macdaily-{}: {}{}system{}: command not found'.format(self.cmd, red_bg, flash, reset), file=sys.stderr)
            text = ('macdaily-{}: {}system{}: '
                    "you may add `softwareupdate' to PATH through the following command -- "
                    """`{}echo 'export PATH="/usr/sbin:$PATH"' >> ~/.bash_profile{}'""".format(self.cmd, red, reset, bold, reset))
            print_term(text, self._file, redirect=self._qflag)
        return flag

    def _loc_exec(self):
        self._exec = {self._var__exec_path}
        del self._var__exec_path

    @abc.abstractmethod
    def _check_list(self, path):
        self._var__rcmd_pkgs = set()  # pylint: disable=attribute-defined-outside-init
        self._var__norm_pkgs = set()  # pylint: disable=attribute-defined-outside-init

    def _check_pkgs(self, path):
        self._check_list(path)
        text = 'Checking existence of specified packages'
        print_info(text, self._file, redirect=self._vflag)

        _rcmd_pkgs = list()
        _norm_pkgs = list()
        _lost_pkgs = list()
        for package in self._packages:
            if package in self._var__rcmd_pkgs:
                _rcmd_pkgs.append(package)
            elif package in self._var__norm_pkgs:
                _norm_pkgs.append(package)
            else:
                _lost_pkgs.append(package)
        self._lost.extend(_lost_pkgs)

        self._var__real_pkgs = self._var__rcmd_pkgs | self._var__norm_pkgs  # pylint: disable=attribute-defined-outside-init
        self._var__lost_pkgs = set(_lost_pkgs)  # pylint: disable=attribute-defined-outside-init
        self._var__rcmd_pkgs = set(_rcmd_pkgs)  # pylint: disable=attribute-defined-outside-init
        self._var__norm_pkgs = set(_norm_pkgs)  # pylint: disable=attribute-defined-outside-init
        self._var__temp_pkgs = self._var__rcmd_pkgs | self._var__norm_pkgs  # pylint: disable=attribute-defined-outside-init



# -*- coding: utf-8 -*-

import abc
import contextlib
import glob
import os
import re
import shutil
import sys
import traceback

from macdaily.cls.command import Command
from macdaily.util.compat import pathlib, subprocess
from macdaily.util.const.term import (bold, flash, green, purple_bg, red, red_bg, reset, under,
                                      yellow)
from macdaily.util.tools.get import get_input
from macdaily.util.tools.make import make_stderr
from macdaily.util.tools.misc import date
from macdaily.util.tools.print import print_info, print_scpt, print_term, print_text
from macdaily.util.tools.script import run, sudo


class BrewCommand(Command):

    @property
    def mode(self):
        return 'brew'

    @property
    def name(self):
        return 'Homebrew'

    @property
    def desc(self):
        return ('Homebrew formula', 'Homebrew formulae')

    def _check_exec(self):
        self._var__exec_path = shutil.which('brew')
        flag = (self._var__exec_path is not None)
        if not flag:
            print('macdaily-{}: {}{}brew{}: command not found'.format(self.cmd, red_bg, flash, reset), file=sys.stderr)
            text = ('macdaily-{}: {}brew{}: you may find Homebrew on '
                    '{}{}https://brew.sh{}, or install Homebrew through following command -- '
                    '`{}/usr/bin/ruby -e "$(curl -fsSL '
                    """https://raw.githubusercontent.com/Homebrew/install/master/install)"{}'""".format(self.cmd, red, reset, purple_bg, under, reset, bold, reset))
            print_term(text, self._file, redirect=self._qflag)
        return flag

    @abc.abstractmethod
    def _parse_args(self, namespace):
        super()._parse_args(namespace)
        self._force = namespace.get('force', False)  # pylint: disable=attribute-defined-outside-init
        self._merge = namespace.get('merge', False)  # pylint: disable=attribute-defined-outside-init
        self._no_cleanup = namespace.get('no_cleanup', False)  # pylint: disable=attribute-defined-outside-init
        self._verbose = namespace.get('verbose', False)  # pylint: disable=attribute-defined-outside-init

    def _loc_exec(self):
        self._exec = {self._var__exec_path}
        del self._var__exec_path

    def _check_pkgs(self, path):
        text = 'Listing installed {}'.format(self.desc[1])
        print_info(text, self._file, redirect=self._vflag)

        argv = [path, 'list']
        args = ' '.join(argv)
        print_scpt(args, self._file, redirect=self._vflag)
        with open(self._file, 'a') as file:
            file.write('Script started on {}\n'.format(date()))
            file.write('command: {!r}\n'.format(args))

        try:
            proc = subprocess.check_output(argv, stderr=make_stderr(self._vflag))
        except subprocess.CalledProcessError:
            print_text(traceback.format_exc(), self._file, redirect=self._vflag)
            _real_pkgs = set()
        else:
            context = proc.decode()
            _real_pkgs = set(context.split())
            print_text(context, self._file, redirect=self._vflag)
        finally:
            with open(self._file, 'a') as file:
                file.write('Script done on {}\n'.format(date()))

        text = 'Checking existence of specified packages'
        print_info(text, self._file, redirect=self._vflag)

        _temp_pkgs = list()
        _lost_pkgs = list()
        for package in self._packages:
            if package in _real_pkgs:
                _temp_pkgs.append(package)
            else:
                _lost_pkgs.append(package)
        self._lost.extend(_lost_pkgs)

        self._var__real_pkgs = set(_real_pkgs)  # pylint: disable=attribute-defined-outside-init
        self._var__lost_pkgs = set(_lost_pkgs)  # pylint: disable=attribute-defined-outside-init
        self._var__temp_pkgs = set(_temp_pkgs)  # pylint: disable=attribute-defined-outside-init

    def _proc_renew(self, path):
        text = 'Updating Homebrew database'
        print_info(text, self._file, redirect=self._qflag)

        argv = [path, 'update']
        if self._force:
            argv.append('--force')
        if self._merge:
            argv.append('--merge')
        if self._quiet:
            argv.append('--quiet')
        if self._verbose:
            argv.append('--verbose')
        print_scpt(' '.join(argv), self._file, redirect=self._qflag)
        run(argv, self._file, redirect=self._qflag)

    def _proc_fixmissing(self, path):
        text = 'Checking broken {} dependencies'.format(self.desc[0])
        print_info(text, self._file, redirect=self._qflag)

        def _proc_check():
            argv = [path, 'missing', '--hide={!r}'.format(",".join(self._ignore))]
            args = ' '.join(argv)
            print_scpt(args, self._file, redirect=self._vflag)
            with open(self._file, 'a') as file:
                file.write('Script started on {}\n'.format(date()))
                file.write('command: {!r}\n'.format(args))

            _deps_pkgs = list()
            try:  # brew missing exits with a non-zero status if any formulae are missing dependencies
                proc = subprocess.run(argv, stdout=subprocess.PIPE, stderr=make_stderr(self._vflag))
            except subprocess.SubprocessError:
                print_text(traceback.format_exc(), self._file, redirect=self._vflag)
            else:
                context = proc.stdout.decode()
                print_text(context, self._file, redirect=self._vflag)

                for line in filter(None, context.strip().splitlines()):
                    _deps_pkgs.extend(line.split()[1:])
            finally:
                with open(self._file, 'a') as file:
                    file.write('Script done on {}\n'.format(date()))
            return set(_deps_pkgs)

        def _proc_confirm():
            pkgs = '{}, {}'.format(reset, bold).join(_deps_pkgs)
            text = 'macdaily-{}: {}brew{}: found broken dependencies: {}{}{}'.format(self.cmd, yellow, reset, bold, pkgs, reset)
            print_term(text, self._file, redirect=self._qflag)
            if self._yes or self._quiet:
                return True
            while True:
                ans = get_input(self._confirm, 'Would you like to reinstall?',
                                prefix='Found broken dependencies: {}.\n\n'.format(", ".join(_deps_pkgs)),
                                suffix=' ({}y{}/{}N{}) '.format(green, reset, red, reset))
                if re.match(r'[yY]', ans):
                    return True
                if re.match(r'[nN]', ans):
                    return False
                print('Invalid input.', file=sys.stderr)

        _deps_pkgs = _proc_check() - self._ignore
        if not _deps_pkgs:
            text = 'macdaily-{}: {}brew{}: no broken dependencies'.format(self.cmd, green, reset)
            print_term(text, self._file, redirect=self._qflag)
            return

        text = 'Fixing broken {} dependencies'.format(self.desc[0])
        print_info(text, self._file, redirect=self._qflag)

        if _proc_confirm():
            argv = [path, 'reinstall']
            if self._quiet:
                argv.append('--quiet')
            if self._verbose:
                argv.append('--verbose')
            argv.append('')

            _done_pkgs = set()
            while _deps_pkgs:
                for package in _deps_pkgs:
                    argv[-1] = package
                    print_scpt(' '.join(argv), self._file, redirect=self._qflag)
                    if not run(argv, self._file, redirect=self._qflag,
                               timeout=self._timeout, verbose=self._vflag):
                        with contextlib.suppress(ValueError):
                            self._pkgs.remove(package)
                _done_pkgs |= _deps_pkgs
                _deps_pkgs = _proc_check() - _done_pkgs - self._ignore

            text = 'macdaily-{}: {}brew{}: all broken dependencies fixed'.format(self.cmd, green, reset)
        else:
            text = 'macdaily-{}: {}brew{}: all broken dependencies remain'.format(self.cmd, red, reset)
        print_term(text, self._file, redirect=self._qflag)

    def _proc_cleanup(self):
        if self._no_cleanup:
            return

        text = 'Pruning caches and archives'
        print_info(text, self._file, redirect=self._qflag)

        argv = ['brew', 'cleanup']
        if self._verbose:
            argv.append('--verbose')
        if self._quiet:
            argv.append('--quiet')
        args = ' '.join(argv)
        print_scpt(args, self._file, redirect=self._qflag)

        flag = (not os.path.isdir(self._disk_dir))
        for path in self._exec:
            logs = os.path.expanduser('~/Library/Logs/Homebrew/')
            if os.path.isdir(logs):
                argv = [path, 'logs', 'cleanup']
                if self._verbose:
                    argv.append('--verbose')
                if self._quiet:
                    argv.append('--quiet')
                args = ' '.join(argv)
                print_scpt(args, self._file, redirect=self._vflag)

                argv = ['rm', '-rf']
                if self._verbose:
                    argv.append('-v')
                argv.append(logs)
                sudo(argv, self._file, self._password,
                     redirect=self._qflag, verbose=self._vflag)

            # if external disk not attached
            if flag:
                continue

            argv = [path, '--cache']
            args = ' '.join(argv)
            print_scpt(args, self._file, redirect=self._vflag)
            with open(self._file, 'a') as file:
                file.write('Script started on {}\n'.format(date()))
                file.write('command: {!r}\n'.format(args))

            fail = False
            try:
                proc = subprocess.check_output(argv, stderr=make_stderr(self._vflag))
            except subprocess.CalledProcessError:
                print_text(traceback.format_exc(), self._file, redirect=self._vflag)
                fail = True
            else:
                context = proc.decode()
                print_text(context, self._file, redirect=self._vflag)
            finally:
                with open(self._file, 'a') as file:
                    file.write('Script done on {}\n'.format(date()))
            if fail:
                continue

            cache = context.strip()
            if os.path.isdir(cache):
                argv = [path, 'caches', 'archive']
                if self._verbose:
                    argv.append('--verbose')
                if self._quiet:
                    argv.append('--quiet')
                args = ' '.join(argv)
                print_scpt(args, self._file, redirect=self._qflag)

                def _move(root, stem):
                    arch = os.path.join(self._disk_dir, stem)
                    pathlib.Path(arch).mkdir(parents=True, exist_ok=True)

                    file_list = list()
                    for name in os.listdir(root):
                        path = os.path.join(root, name)
                        if os.path.isdir(path):
                            if name != 'Cask':
                                file_list.extend(_move(path, os.path.join(stem, name)))
                        elif os.path.splitext(name)[1] != '.incomplete' and path not in cask_list:  # pylint: disable=cell-var-from-loop
                            try:
                                shutil.move(path, os.path.join(arch, name))
                            except (shutil.Error, FileExistsError):
                                os.remove(path)
                            # with contextlib.suppress(shutil.Error, FileExistsError):
                            #     shutil.move(path, os.path.join(arch, name))
                            file_list.append(path)
                    return file_list

                cask_list = [os.path.realpath(name) for name in glob.glob(os.path.join(cache, 'Cask/*'))]
                file_list = _move(cache, 'Homebrew')
                print_text(os.linesep.join(sorted(file_list)), self._file, redirect=self._vflag)

        if flag:
            text = ('macdaily-{}: {}brew{}: '
                    'archive directory {}{}{} not found'.format(self.cmd, yellow, reset, bold, self._disk_dir, reset))
            print_term(text, self._file, redirect=self._vflag)



# -*- coding: utf-8 -*-

import abc
import copy
import os
import re
import sys

from macdaily.util.compat import subprocess
from macdaily.util.const.macro import NODE, PYTHON2, PYTHON3, USR
from macdaily.util.const.term import bold, green, red, reset, yellow
from macdaily.util.tools.get import get_input, get_logfile
from macdaily.util.tools.print import print_info, print_term
from macdaily.util.tools.script import sudo


class Command(metaclass=abc.ABCMeta):
    """Base command.

    Process
    ~~~~~~~

    1. check executable
        1. if none exits, exit
        2. else continue
    2. parse options and packages
        1. merge package specification in options
        2. extract command line options
        3. if no package specifications and ``all`` flag NOT set, exit
        4. else continue
    3. locate executables
    4. run command-specified processors
        1. for each executable
            1. command-specified logging process (optional)
                1. fetch packages for main process
                2. if found package specifications, provide trivial did-you-mean function
                3. else continue
            2. ask for comfirmation on main process
                1. if cancelled, exit
                2. else continue
            3. command-specified main process
                1. run main process for each package
                2. run checkout process (optional)
        2. run cleanup process (optional)

    Properties
    ~~~~~~~~~~

    - ``cmd`` -- ``str``, command type
    - ``act`` -- ``tuple<str>``, command actions
        0. verb
        1. verb (past participle)
        2. adjective
    - ``job`` -- ``tuple<str>``, command jobs
        0. noun (singular)
        1. noun (plural)
    - ``name`` -- ``str``, command name (full name)
    - ``mode`` -- ``str``, command mode (acronym)
    - ``time`` -- ``float`` / ``None``, Homebrew renew timestamp
    - ``desc`` -- ``tuple<str>``, command description
        0. singular
        1. plural
    - ``packages`` -- ``set<str>``, process succeeded packages
    - ``ignored`` -- ``set<str>``, ignored packages (specified in options)
    - ``failed`` -- ``set<str>``, process failed packages
    - ``notfound`` -- ``set<str>``, unknown packages (not found in registry)

    """
    @property
    @abc.abstractmethod
    def cmd(self):
        return NotImplemented

    @property
    @abc.abstractmethod
    def act(self):
        """verb, past participle, adjective"""
        return (NotImplemented, NotImplemented, NotImplemented)

    @property
    @abc.abstractmethod
    def job(self):
        """noun singular, noun plural"""
        return (NotImplemented, NotImplemented)

    @property
    @abc.abstractmethod
    def name(self):
        return NotImplemented

    @property
    @abc.abstractmethod
    def mode(self):
        return NotImplemented

    @property
    def time(self):
        return self._brew_renew

    @property
    @abc.abstractmethod
    def desc(self):
        """singular, plural"""
        return (NotImplemented, NotImplemented)

    @property
    def packages(self):
        return set(self._pkgs)

    @property
    def ignored(self):
        return set(self._ilst)

    @property
    def failed(self):
        return set(self._fail)

    @property
    def notfound(self):
        return set(self._lost)

    def __init__(self, namespace, filename, timeout, confirm,
                 askpass, password, disk_dir, brew_renew=None):
        """Initialisation.

        Args:

        - ``namespace`` -- ``dict``, converted argparse.Namespace
        - ``filename`` -- ``str``, real path of log file
        - ``timeout`` -- ``int``, timeout interval for main process
        - ``confirm`` -- ``str``, path to ``macdaily-confirm``
        - ``askpass`` -- ``str``, path to ``macdaily-askpass``
        - ``password`` --  ``str``/``bytes``, sudo password
        - ``disk_dir`` -- ``str``, real root path of archive directory
        - ``brew_renew`` -- ``float``, Homebrew renew timestamp

        """
        self._qflag = namespace.get('quiet', False)
        self._vflag = self._qflag or (not namespace.get('verbose', False))

        text = 'Running {} command for {}'.format(self.cmd, self.mode)
        print_info(text, filename, redirect=self._qflag)

        # assign members
        self._file = filename
        self._timeout = timeout
        self._confirm = confirm
        self._askpass = askpass
        self._password = password
        self._disk_dir = disk_dir
        self._brew_renew = brew_renew

        # exit if no executable found
        if self._check_exec():
            # mainloop process
            if self._pkg_args(namespace):
                self._loc_exec()
                self._run_proc()
            else:
                text = ('macdaily-{}: {}{}{}: '
                        'no {}{}{} to {}'.format(self.cmd, yellow, self.mode, reset, bold, self.desc[1], reset, self.act[0]))
                print_term(text, filename, redirect=self._qflag)
        else:
            self._pkgs = list()
            self._fail = list()
            self._lost = list()
            self._ilst = list()

        # remove temp vars
        [delattr(self, attr) for attr in filter(lambda s: s.startswith('_var_'), dir(self))]  # pylint: disable=expression-not-assigned

    @abc.abstractmethod
    def _check_exec(self):
        """Return if executable found."""
        return True

    def _pkg_args(self, namespace):
        """Return if there's packages for main process."""
        self._merge_packages(namespace)
        self._parse_args(namespace)

        self._pkgs = list()
        self._fail = list()
        self._lost = list()
        self._ilst = copy.copy(self._ignore)

        return bool(self._packages or self._all)

    def _merge_packages(self, namespace):
        ilst_pkg = list()
        temp_pkg = list()
        args_pkg = namespace.get('packages', list())
        for pkgs in args_pkg:
            if isinstance(pkgs, str):
                pkgs = filter(None, pkgs.split(','))  # pylint: disable=filter-builtin-not-iterating
            for item in map(lambda s: s.split(','), pkgs):
                for package in map(lambda s: s.strip(), item):
                    if package.startswith('!'):
                        ilst_pkg.append(package[1:])
                    else:
                        temp_pkg.append(package)
        self._ignore = set(ilst_pkg)  # pylint: disable=attribute-defined-outside-init
        self._packages = set(temp_pkg)  # pylint: disable=attribute-defined-outside-init

    @abc.abstractmethod
    def _parse_args(self, namespace):
        self._all = namespace.get('all', False)  # pylint: disable=attribute-defined-outside-init
        self._quiet = namespace.get('quiet', False)  # pylint: disable=attribute-defined-outside-init
        self._verbose = namespace.get('verbose', False)  # pylint: disable=attribute-defined-outside-init
        self._yes = namespace.get('yes', False)  # pylint: disable=attribute-defined-outside-init

    @abc.abstractmethod
    def _loc_exec(self):
        self._exec = set()

    @abc.abstractmethod
    def _run_proc(self):
        self._pkgs = list()
        self._fail = list()
        self._lost = list()
        for path in self._exec:
            self._var__lost_pkgs = set()
            self._var__real_pkgs = set()
            self._var__temp_pkgs = set()
            self._proc_fixmissing(path)
        self._proc_cleanup()

    def _check_confirm(self, path):
        def chown(flag=False):
            if not flag or self.mode != 'brew':
                return flag

            def _cross_check(packages, aliases):
                for formula in aliases:
                    if formula in packages:
                        return True
                return False

            if _cross_check(self._var__temp_pkgs, NODE):
                sudo(['chown', '-R', USR, '/usr/local/lib/node_modules'], get_logfile(), self._password,
                     timeout=self._timeout, redirect=True, verbose=False)

            if _cross_check(self._var__temp_pkgs, PYTHON2):
                version = subprocess.check_output([path, 'info', 'python@2']).splitlines()[0].split()[2][:3].decode()
                sudo(['chown', '-R', USR, '/usr/local/lib/python{}'.format(version)], get_logfile(), self._password,
                     timeout=self._timeout, redirect=True, verbose=False)

            if _cross_check(self._var__temp_pkgs, PYTHON3):
                version = subprocess.check_output([path, 'info', 'python']).splitlines()[0].split()[2][:3].decode()
                sudo(['chown', '-R', USR, '/usr/local/lib/python{}'.format(version)], get_logfile(), self._password,
                     timeout=self._timeout, redirect=True, verbose=False)

            return flag

        self._var__temp_pkgs -= self._ignore
        if not self._var__temp_pkgs:
            text = 'macdaily-{}: {}{}{}: no {}{}{} to {}'.format(self.cmd, green, self.mode, reset, bold, self.desc[1], reset, self.act[0])
            print_term(text, self._file, redirect=self._qflag)
            return chown(flag=True)

        job = self.job[1] if self._var__temp_pkgs else self.job[0]
        bold_pkgs = '{}, {}'.format(reset, bold).join(self._var__temp_pkgs)
        text = ('macdaily-{}: {}{}{}: '
                '{} {} available for {}{}{}'.format(self.cmd, green, self.mode, reset, self.desc[0], job, bold, bold_pkgs, reset))
        print_term(text, self._file, redirect=self._qflag)
        if self._yes or self._quiet:
            return chown(flag=True)
        while True:
            ans = get_input(self._confirm, 'Would you like to {}?'.format(self.act[0]),
                            prefix='{} {} available for {}.\n\n'.format(self.desc[0], job, ", ".join(self._var__temp_pkgs)),
                            suffix=' ({}y{}/{}N{}) '.format(green, reset, red, reset))
            if re.match(r'[yY]', ans):
                return chown(flag=True)
            if re.match(r'[nN]', ans):
                text = ('macdaily-{}: {}{}{}: '
                        '{} {} postponed due to user cancellation'.format(self.cmd, yellow, self.mode, reset, self.desc[0], job))
                print_term(text, self._file, redirect=self._qflag)
                return chown(flag=False)
            print('Invalid input.', file=sys.stderr)

    def _did_you_mean(self):
        for package in self._var__lost_pkgs:
            pattern = r'.*{}.*'.format(package)
            matches = '{}, {}'.format(reset, bold).join(filter(lambda s: re.match(pattern, s, re.IGNORECASE),  # pylint: disable=cell-var-from-loop
                                                     self._var__real_pkgs))
            print('macdaily-{}: {}{}{}: '
                  'no available {} with the name {}{!r}{}'.format(self.cmd, red, self.mode, reset, self.desc[0], bold, package, reset), file=sys.stderr)
            if matches:
                text = ('macdaily-{}: {}{}{}: '
                        'did you mean any of the following {}: {}{}{}?'.format(self.cmd, yellow, self.mode, reset, self.desc[1], bold, matches, reset))
                print_term(text, self._file, redirect=self._qflag)
        del self._var__lost_pkgs
        del self._var__real_pkgs

    def _proc_fixmissing(self, path):
        pass

    def _proc_cleanup(self):
        pass



# -*- coding: utf-8 -*-

import argparse
import sys

from macdaily.util.const.macro import STR_CONFIG
from macdaily.util.const.macro import VERSION as __version__
from macdaily.util.const.term import bold, reset, under


def get_config_parser():
    #######################################################
    # Config CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    parser = argparse.ArgumentParser(prog='macdaily-config',
                                     description='MacDaily Runtime Configuration Helper',
                                     usage='macdaily config [options] <key> <value> ...',
                                     epilog=STR_CONFIG)
    parser.add_argument('-V', '--version', action='version', version=__version__)

    spec_group = parser.add_argument_group(title='specification arguments')
    spec_group.add_argument('-a', '--add', action='store_true',
                            help=('adds a new line to the option without altering any '
                                  'existing values [requires {}ConfigUpdater{}]'.format(under, reset)))
    spec_group.add_argument('-g', '--get', action='store_true',
                            help='get the value for a given key')
    spec_group.add_argument('-u', '--unset', action='store_true',
                            help=('remove the line matching the key from config file '
                                  '[requires {}ConfigUpdater{}]'.format(under, reset)))
    spec_group.add_argument('-i', '--interactive', action='store_true',
                            help='enter interactive configuration mode')
    spec_group.add_argument('-l', '--list', action='store_true',
                            help='list all variables set in config file, along with their values')

    genl_group = parser.add_argument_group(title='general arguments')
    genl_group.add_argument('key', action='store', nargs='?', help='a given key')
    genl_group.add_argument('value', action='store', nargs='?', help='the value for a given key')
    genl_group.add_argument('-q', '--quiet', action='store_true',
                            help='run in quiet mode, with no output information')
    genl_group.add_argument('-v', '--verbose', action='store_true',
                            help='run in verbose mode, with detailed output information')

    ctrl_group = parser.add_argument_group(title='control arguments',
                                           description='options used to set true or false')
    ctrl_group.add_argument('-T', '--true', action='store_true',
                            help="set the value for a given key to `{}true{}'".format(bold, reset))
    ctrl_group.add_argument('-F', '--false', action='store_true',
                            help="set the value for a given key to `{}false{}'".format(bold, reset))

    return parser


def parse_args(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    # main parser process
    main_parser = get_config_parser()
    main_args = main_parser.parse_args(argv or ['--help'])

    return main_args



# -*- coding: utf-8 -*-
"""utility macros"""

import getpass
import os
import pwd
import shutil
import sys

from macdaily.util.compat import pathlib

# version string
VERSION = '2019.4.7.post1'

# terminal commands
PYTHON = sys.executable         # Python version
PROGRAM = ' '.join(sys.argv)    # arguments

###########################################################
# Environment Variables
###########################################################

# boolean mappings
BOOLEAN_STATES = {'1': True, '0': False,
                  'yes': True, 'no': False,
                  'true': True, 'false': False,
                  'on': True, 'off': False}

# DEVMODE flag
DEVMODE = BOOLEAN_STATES.get(os.getenv('MACDAILY_DEVMODE', 'false').strip().lower(), False)

###########################################################
# Miscellaneous Constants
###########################################################

# ansi escape pattern
# from http://www.umich.edu/~archive/apple2/misc/programmers/vt100.codes.txt
ANSI = r'([\x1B\x9B][\[\]\(\)#;?]*(?:(?:(?:[a-zA-Z0-9]*(?:;[-a-zA-Z0-9\\/#&.:=?%@~_]*)*)?\x07)|(?:(?:\d{1,4}(?:;\d{0,4})*)?[0-9A-PR-TZcf-ntqry=><~])))'

# script utilities
SCRIPT = shutil.which('script')
UNBUFFER = shutil.which('unbuffer')

# environment macros
ROOT = str(pathlib.Path(__file__).resolve().parents[2])
SHELL = os.getenv('SHELL', shutil.which('sh'))

# user information
USR = getpass.getuser()
PWD = pwd.getpwnam(USR)
USER = PWD.pw_gecos
PASS = os.getenv('SUDO_PASSWORD', PWD.pw_passwd)

# emoji mappings
ORIG_BEER = '🍺'
RESP_BEER = 'ð\x9f\x8dº'

# homebrew aliases
NODE = {'node', 'node.js', 'node@11', 'nodejs', 'npm'}
PYTHON2 = {'python@2', 'python2'}
PYTHON3 = {'python', 'python3', 'python@3'}

###########################################################
# Command Mappings
###########################################################

# available commands
COMMANDS = '''\
MacDaily available commands & corresponding subsidiaries:
    archive
    bundle          dump, load
    cleanup         brew, cask, npm, pip
    commands
    config
    dependency      brew, pip
    help
    install         apm, brew, cask, gem, mas, npm, pip, system
    launch
    logging         apm, app, brew, cask, gem, mas, npm, pip, tap
    postinstall
    reinstall       brew, cask
    uninstall       brew, cask, pip
    update          apm, brew, cask, gem, mas, npm, pip, system
'''

# command aliases
LNK_ARCHIVE = set()
LNK_BUNDLE = set()
LNK_CLEANUP = {'clean'}
LNK_COMMANDS = set()
LNK_CONFIG = {'cfg'}
LNK_DEPENDENCY = {'deps', 'dp'}
LNK_HELP = {'doc', 'man'}
LNK_INSTALL = {'i'}
LNK_LAUNCH = {'init'}
LNK_LOGGING = {'log'}
LNK_POSTINSTALL = {'post', 'ps'}
LNK_REINSTALL = {'re'}
LNK_UNINSTALL = {'un', 'unlink', 'remove', 'rm', 'r'}
LNK_UPDATE = {'up', 'upgrade'}

# aliases strings
STR_ARCHIVE = None
STR_BUNDLE = None
STR_CLEANUP = "aliases: {}".format(', '.join(sorted(LNK_CLEANUP)))
STR_CONFIG = "aliases: {}".format(', '.join(sorted(LNK_CONFIG)))
STR_DEPENDENCY = "aliases: {}".format(', '.join(sorted(LNK_DEPENDENCY)))
STR_HELP = "aliases: {}".format(', '.join(sorted(LNK_HELP)))
STR_INSTALL = "aliases: {}".format(', '.join(sorted(LNK_INSTALL)))
STR_LAUNCH = "aliases: {}".format(', '.join(sorted(LNK_LAUNCH)))
STR_LOGGING = "aliases: {}".format(', '.join(sorted(LNK_LOGGING)))
STR_POSTINSTALL = "aliases: {}".format(', '.join(sorted(LNK_POSTINSTALL)))
STR_REINSTALL = "aliases: {}".format(', '.join(sorted(LNK_REINSTALL)))
STR_UNINSTALL = "aliases: {}".format(', '.join(sorted(LNK_UNINSTALL)))
STR_UPDATE = "aliases: {}".format(', '.join(sorted(LNK_UPDATE)))

# command mappings
MAP_ARCHIVE = {'archive'} | LNK_ARCHIVE
MAP_BUNDLE = {'bundle'} | LNK_BUNDLE
MAP_CLEANUP = {'cleanup'} | LNK_CLEANUP
MAP_COMMANDS = {'commands'} | LNK_COMMANDS
MAP_CONFIG = {'config'} | LNK_CONFIG
MAP_DEPENDENCY = {'dependency'} | LNK_DEPENDENCY
MAP_HELP = {'help'} | LNK_HELP
MAP_INSTALL = {'install'} | LNK_INSTALL
MAP_LAUNCH = {'launch'} | LNK_LAUNCH
MAP_LOGGING = {'logging'} | LNK_LOGGING
MAP_MAGIC = {'magic', 'moo', 'whoop-de-doo'}
MAP_POSTINSTALL = {'postinstall'} | LNK_POSTINSTALL
MAP_REINSTALL = {'reinstall'} | LNK_REINSTALL
MAP_UNINSTALL = {'uninstall'} | LNK_UNINSTALL
MAP_UPDATE = {'update'} | LNK_UPDATE
MAP_ALL = (MAP_ARCHIVE | MAP_BUNDLE | MAP_CLEANUP | MAP_COMMANDS | MAP_CONFIG | MAP_DEPENDENCY | MAP_HELP |
           MAP_INSTALL | MAP_LAUNCH | MAP_LOGGING | MAP_POSTINSTALL | MAP_REINSTALL | MAP_UNINSTALL | MAP_UPDATE)

# mode aliases
LNK_APM = {'atom'}
LNK_APP = {'macos', 'application'}
LNK_BREW = {'homebrew'}
LNK_CASK = {'caskroom', 'brew-cask'}
LNK_GEM = {'ruby', 'rubygems'}
LNK_MAS = {'mac', 'appstore', 'app-store', 'mac-app-store'}
LNK_NPM = {'node', 'node.js'}
LNK_PIP = {'python', 'cpython', 'pypy'}
LNK_SYSTEM = {'software', 'softwareupdate'}
LNK_TAP = {'brew-tap'}

# aliases strings
STR_APM = "aliases: {}".format(', '.join(sorted(LNK_APM)))
STR_APP = "aliases: {}".format(', '.join(sorted(LNK_APP)))
STR_BREW = "aliases: {}".format(', '.join(sorted(LNK_BREW)))
STR_CASK = "aliases: {}".format(', '.join(sorted(LNK_CASK)))
STR_GEM = "aliases: {}".format(', '.join(sorted(LNK_GEM)))
STR_MAS = "aliases: {}".format(', '.join(sorted(LNK_MAS)))
STR_NPM = "aliases: {}".format(', '.join(sorted(LNK_NPM)))
STR_PIP = "aliases: {}".format(', '.join(sorted(LNK_PIP)))
STR_SYSTEM = "aliases: {}".format(', '.join(sorted(LNK_SYSTEM)))
STR_TAP = "aliases: {}".format(', '.join(sorted(LNK_TAP)))

# mode mappings
MAP_APM = {'apm'} | LNK_APM
MAP_APP = {'app'} | LNK_APP
MAP_BREW = {'brew'} | LNK_BREW
MAP_CASK = {'cask'} | LNK_CASK
MAP_GEM = {'gem'} | LNK_GEM
MAP_MAS = {'mas'} | LNK_MAS
MAP_NPM = {'npm'} | LNK_NPM
MAP_PIP = {'pip'} | LNK_PIP
MAP_SYSTEM = {'system'} | LNK_SYSTEM
MAP_TAP = {'tap'} | LNK_TAP

# mode dictionary
MAP_DICT = dict()
for mode, link in (('apm', MAP_APM),
                   ('app', MAP_APP),
                   ('brew', MAP_BREW),
                   ('cask', MAP_CASK),
                   ('gem', MAP_GEM),
                   ('mas', MAP_MAS),
                   ('npm', MAP_NPM),
                   ('pip', MAP_PIP),
                   ('system', MAP_SYSTEM),
                   ('tap', MAP_TAP)):
    for alias in link:
        MAP_DICT[alias] = mode

# command subsidiaries
CMD_ARCHIVE = set()
CMD_BUNDLE = {'dump', 'load'}
CMD_CLEANUP = MAP_BREW | MAP_CASK | MAP_NPM | MAP_PIP
CMD_COMMANDS = set()
CMD_CONFIG = set()
CMD_DEPENDENCY = MAP_BREW | MAP_PIP
CMD_HELP = set()
CMD_INSTALL = MAP_APM | MAP_BREW | MAP_CASK | MAP_GEM | MAP_MAS | MAP_NPM | MAP_PIP | MAP_SYSTEM
CMD_LAUNCH = set()
CMD_LOGGING = MAP_APM | MAP_APP | MAP_BREW | MAP_CASK | MAP_GEM | MAP_MAS | MAP_NPM | MAP_PIP | MAP_TAP
CMD_POSTINSTALL = set()
CMD_REINSTALL = MAP_BREW | MAP_CASK
CMD_UNINSTALL = MAP_BREW | MAP_CASK | MAP_PIP
CMD_UPDATE = MAP_APM | MAP_BREW | MAP_CASK | MAP_GEM | MAP_MAS | MAP_NPM | MAP_PIP | MAP_SYSTEM



# -*- coding: utf-8 -*-

import datetime
import os
import sys
import traceback
import uuid

from macdaily.cli.logging import parse_args
from macdaily.cls.logging.apm import ApmLogging  # pylint: disable=unused-import
from macdaily.cls.logging.app import AppLogging  # pylint: disable=unused-import
from macdaily.cls.logging.brew import BrewLogging  # pylint: disable=unused-import
from macdaily.cls.logging.cask import CaskLogging  # pylint: disable=unused-import
from macdaily.cls.logging.gem import GemLogging  # pylint: disable=unused-import
from macdaily.cls.logging.mas import MasLogging  # pylint: disable=unused-import
from macdaily.cls.logging.npm import NpmLogging  # pylint: disable=unused-import
from macdaily.cls.logging.pip import PipLogging  # pylint: disable=unused-import
from macdaily.cls.logging.tap import TapLogging  # pylint: disable=unused-import
from macdaily.cmd.archive import make_archive, make_storage
from macdaily.cmd.config import parse_config
from macdaily.util.compat import pathlib, subprocess
from macdaily.util.const.macro import VERSION as __version__
from macdaily.util.const.term import bold, green, purple, red, reset, under, yellow
from macdaily.util.tools.deco import beholder
from macdaily.util.tools.get import get_pass, get_logfile
from macdaily.util.tools.make import make_namespace
from macdaily.util.tools.misc import record
from macdaily.util.tools.print import print_misc, print_term, print_text


@beholder
def logging(argv=None):
    # parse args & set context redirection flags
    args = parse_args(argv)
    quiet = args.quiet
    verbose = (args.quiet or not args.verbose)

    # parse config & change environ
    config = parse_config(quiet, verbose)
    os.environ['SUDO_ASKPASS'] = config['Miscellaneous']['askpass']

    # fetch current time
    today = datetime.datetime.today()
    logdate = datetime.date.strftime(today, r'%y%m%d')
    logtime = datetime.date.strftime(today, r'%H%M%S')

    # prepare command paras
    confirm = config['Miscellaneous']['confirm']
    askpass = config['Miscellaneous']['askpass']
    timeout = config['Miscellaneous']['limit']
    disk_dir = config['Path']['arcdir']
    brew_renew = None

    # record program status
    text_record = '{}{}|🚨|{} {}Running MacDaily version {}{}'.format(bold, green, reset, bold, __version__, reset)
    print_term(text_record, get_logfile(), redirect=quiet)
    record(get_logfile(), args, today, config, redirect=verbose)

    # ask for password
    text_askpass = '{}{}|🔑|{} {}Your {}sudo{}{} password may be necessary{}'.format(bold, purple, reset, bold, under, reset, bold, reset)
    print_term(text_askpass, get_logfile(), redirect=quiet)
    password = get_pass(askpass)

    cmd_list = list()
    log_list = list()
    file_list = list()
    for mode in {'apm', 'app', 'brew', 'cask', 'gem', 'mas', 'npm', 'pip', 'tap'}:
        # mkdir for logs
        logpath = pathlib.Path(os.path.join(config['Path']['logdir'], 'logging', mode, logdate))
        logpath.mkdir(parents=True, exist_ok=True)
        filename = os.path.join(logpath, '{}-{!s}.log'.format(logtime, uuid.uuid4()))
        os.environ['MACDAILY_LOGFILE'] = filename

        # redo program status records
        print_term(text_record, filename, redirect=True)
        record(filename, args, today, config, redirect=True)
        print_term(text_askpass, filename, redirect=True)

        # skip disabled commands
        if (not config['Mode'].get(mode, False)) or getattr(args, 'no_{}'.format(mode), False):
            text = 'macdaily-logging: {}{}{}: command disabled'.format(yellow, mode, reset)
            print_term(text, filename, redirect=verbose)
            continue

        # update logging specifications
        namespace = getattr(args, mode, vars(args))

        # check master controlling flags
        if args.quiet:
            namespace['quiet'] = True
        if args.verbose:
            namespace['verbose'] = True
        if args.show_log:
            namespace['show_log'] = True
        if args.no_cleanup:
            namespace['no_cleanup'] = True

        # run command
        cmd_cls = globals()['{}Logging'.format(mode.capitalize())]
        command = cmd_cls(make_namespace(namespace), filename, timeout,
                          confirm, askpass, password, disk_dir, brew_renew)

        # record command
        cmd_list.append(command)
        log_list.append(filename)
        brew_renew = command.time

        if not namespace['no_cleanup']:
            archive = make_archive(config, 'logging/{}'.format(mode), today, zipfile=False,
                                   quiet=quiet, verbose=verbose, logfile=filename)
            file_list.extend(archive)

        if namespace.get('show_log', False):
            try:
                subprocess.check_call(['open', '-a', '/Applications/Utilities/Console.app', filename])
            except subprocess.CalledProcessError:
                print_text(traceback.format_exc(), filename, redirect=verbose)
                print('macdaily: {}logging{}: cannot show log file {!r}'.format(red, reset, filename), file=sys.stderr)

    if not args.no_cleanup:
        storage = make_storage(config, today, quiet=quiet, verbose=verbose, logfile=filename)
        file_list.extend(storage)

    text = '{}{}|📖|{} {}MacDaily report of logging command{}'.format(bold, green, reset, bold, reset)
    print_term(text, get_logfile(), redirect=quiet)
    for file in log_list:
        print_term(text, file, redirect=True)

    for command in cmd_list:
        text = 'Recorded existing {}{}{}{} at {}{}{}'.format(under, command.desc[1], reset, bold, under, command.sample, reset)
        print_misc(text, get_logfile(), redirect=quiet)
        for file in log_list:
            print_misc(text, file, redirect=True)

    if file_list:
        formatted_list = '{}{}, {}'.format(reset, bold, under).join(file_list)
        text = ('Archived following ancient logs: {}{}{}'.format(under, formatted_list, reset))
        print_misc(text, filename, redirect=quiet)

    if len(cmd_list) == 0:  # pylint: disable=len-as-condition
        text = 'macdaily: {}logging{}: no packages recorded'.format(purple, reset)
        print_term(text, get_logfile(), redirect=quiet)
        for file in log_list:
            print_term(text, file, redirect=True)

    mode_lst = [command.mode for command in cmd_list]
    mode_str = ', '.join(mode_lst) if mode_lst else 'none'
    text = ('{}{}|🍺|{} {}MacDaily successfully performed logging process '
            'for {} package managers{}'.format(bold, green, reset, bold, mode_str, reset))
    print_term(text, get_logfile(), redirect=quiet)
    for file in log_list:
        print_term(text, file, redirect=True)


if __name__ == '__main__':
    sys.exit(logging())



# -*- coding: utf-8 -*-

import argparse
import sys

from macdaily.util.const.macro import STR_ARCHIVE
from macdaily.util.const.macro import VERSION as __version__


def get_archive_parser():
    #######################################################
    # Archive CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    parser = argparse.ArgumentParser(prog='macdaily-archive',
                                     description='MacDaily Log Archive Utility',
                                     usage='macdaily archive [options] <path-selection> ...',
                                     epilog=STR_ARCHIVE)
    parser.add_argument('-V', '--version', action='version', version=__version__)

    spec_group = parser.add_argument_group(title='specification arguments')
    spec_group.add_argument('path', nargs='*', metavar='CMD',
                            help=('archive logs of specified command, e.g. archive, cleanup, '
                                  'dependency, logging, postinstall, reinstall, uninstall, update, '
                                  'logging/apm, logging/app, logging/brew, logging/cask, logging/gem, '
                                  'logging/mas, logging/npm, logging/pip and logging/tap'))

    genl_group = parser.add_argument_group(title='general arguments')
    genl_group.add_argument('-a', '--all', action='store_true',
                            help='archive all ancient logs')
    genl_group.add_argument('-n', '--no-storage', action='store_true',
                            help='do not move ancient logs into external hard disk')
    genl_group.add_argument('-q', '--quiet', action='store_true',
                            help='run in quiet mode, with no output information')
    genl_group.add_argument('-v', '--verbose', action='store_true',
                            help='run in verbose mode, with detailed output information')
    genl_group.add_argument('-l', '--show-log', action='store_true',
                            help='open log in Console.app upon completion of command')

    return parser


def parse_args(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    # main parser process
    main_parser = get_archive_parser()
    main_args = main_parser.parse_args(argv or ['--help'])

    return main_args



# -*- coding: utf-8 -*-

import argparse
import re
import sys

from macdaily.api.archive import archive
# from macdaily.api.bundle import bundle
from macdaily.api.cleanup import cleanup
from macdaily.api.config import config
from macdaily.api.dependency import dependency
from macdaily.api.help import help_
from macdaily.api.install import install
from macdaily.api.launch import launch
from macdaily.api.logging import logging
from macdaily.api.postinstall import postinstall
from macdaily.api.reinstall import reinstall
from macdaily.api.uninstall import uninstall
from macdaily.api.update import update
from macdaily.res.magic import whoop_de_doo
from macdaily.util.const.macro import (COMMANDS, MAP_ALL, MAP_ARCHIVE, MAP_BUNDLE, MAP_CLEANUP,
                                       MAP_COMMANDS, MAP_CONFIG, MAP_DEPENDENCY, MAP_HELP,
                                       MAP_INSTALL, MAP_LAUNCH, MAP_LOGGING, MAP_MAGIC,
                                       MAP_POSTINSTALL, MAP_REINSTALL, MAP_UNINSTALL, MAP_UPDATE)
from macdaily.util.const.macro import VERSION as __version__
from macdaily.util.const.term import bold, reset
from macdaily.util.error import CommandNotImplemented
from macdaily.util.tools.deco import beholder
from macdaily.util.tools.print import print_environ


def get_parser():
    parser = argparse.ArgumentParser(prog='MacDaily',
                                     description='macOS Automated Package Manager',
                                     usage='macdaily [options] <command> ...',
                                     epilog='{} This MacDaily has Super Cow Powers.{}'.format(bold, reset))
    parser.add_argument('-V', '--version', action='version', version=__version__)
    parser.add_argument('-E', '--environ', action='store_true',
                        help='show all available environment variables')

    group = parser.add_argument_group('command selection',
                                      'MacDaily provides a friendly CLI workflow for the '
                                      'administrator of macOS to manipulate packages, see '
                                      "`{}macdaily commands{}' for more information".format(bold, reset))
    group.add_argument('command', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    return parser


@beholder
def main():
    # parse args
    parser = get_parser()
    args = parser.parse_args(sys.argv[1:2] or ['--help'])

    # list environs
    if args.environ:
        return print_environ()

    # fetch command & paras
    command = args.command[0].strip().lower()
    options = sys.argv[2:]

    if command in MAP_MAGIC:
        whoop_de_doo()
    elif command in MAP_COMMANDS:
        print(COMMANDS, end='')
    elif command in MAP_ARCHIVE:
        archive(options)
    elif command in MAP_BUNDLE:
        # bundle(options)
        raise CommandNotImplemented
    elif command in MAP_CLEANUP:
        cleanup(options)
    elif command in MAP_CONFIG:
        config(options)
    elif command in MAP_DEPENDENCY:
        dependency(options)
    elif command in MAP_HELP:
        help_(options)
    elif command in MAP_INSTALL:
        install(options)
    elif command in MAP_LAUNCH:
        launch(options)
    elif command in MAP_LOGGING:
        logging(options)
    elif command in MAP_POSTINSTALL:
        postinstall(options)
    elif command in MAP_REINSTALL:
        reinstall(options)
    elif command in MAP_UNINSTALL:
        uninstall(options)
    elif command in MAP_UPDATE:
        update(options)
    else:
        pattern = r'.*{}.*'.format(command)
        matches = "', '".format().join(filter(lambda s: re.match(pattern, s, re.IGNORECASE), MAP_ALL))  # pylint: disable=cell-var-from-loop
        if matches:
            parser.error("argument CMD: invalid choice: {!r} "
                         "(did you mean: '{}')".format(command, matches))
        else:
            parser.error("argument CMD: invalid choice: {!r} "
                         "(choose from '{}')".format(command, ', '.join(sorted(MAP_ALL))))


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-

import os
import random
import shutil

from macdaily.util.compat import subprocess

# useful programmes
FORTUNE = shutil.which('fortune')
COWSAY = shutil.which('cowsay')
LOLCAT = shutil.which('lolcat')

# my own quotes
DB = {
    ("N avpur bs gur ynzo vf gur tenivgnf bs n snepr.\n"
     "\n"
     "-- Wneel Funj --"),
    ("Oyrffrq ner gur zrrx.\n"
     "\n"
     "Naq oyrffrq ner gubfr jub fhssre sbe gur pnhfr bs\n"
     "evtugrbhfarff sbe gurvef vf gur xvatqbz bs urnira."),
    ("Punbf vfa'g n cvg.\n"
     "Punbf vf gur ynqqre.\n"
     "\n"
     "-- Crgle Ornyvfu --"),
    ("Gul xvatqbz pbzr.\n"
     "Gul jvyy or qbar...\n"
     "\n"
     "-- Fureybpx --"),
}


def cowfile():
    try:
        proc = subprocess.check_output([COWSAY, '-l'], stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        return list()

    cows = list()
    for line in proc.strip().splitlines()[1:]:
        cows.extend(line.decode().split())
    return cows


def decode():
    d = {}
    for c in (65, 97):
        for i in range(26):
            d[chr(i+c)] = chr((i+13) % 26 + c)
    return ['echo "{}"'.format("".join([d.get(c, c) for c in s])) for s in DB]


def install(formula):
    BREW = shutil.which('brew')
    if BREW is None:
        return False
    try:
        subprocess.check_call([BREW, 'install', formula],
                              stdout=subprocess.DEVNULL,
                              stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        return False
    return True


def whoop_de_doo():
    s = decode()
    if FORTUNE is None:
        if install('fortune'):
            s.extend(['/usr/local/opt/fortune/bin/fortune' for _ in DB])
    else:
        s.extend([FORTUNE for _ in DB])
    fortune = random.choice(s)

    if COWSAY is None:
        if install('cowsay'):
            exec_list = ['/usr/local/opt/cowsay/bin/cowsay', '/usr/local/opt/cowsay/bin/cowthink']
            cowsay = '{} -f {}'.format(random.choice(exec_list), random.choice(cowfile()))
        else:
            cowsay = 'cat'
    else:
        COWTHINK = shutil.which('cowthink')
        if COWTHINK is None:
            exec_list = [COWSAY]
        else:
            exec_list = [COWSAY, COWTHINK]
        cowsay = '{} -f {}'.format(random.choice(exec_list), random.choice(cowfile()))

    if LOLCAT is None:
        if install('lolcat'):
            lolcat = '/usr/local/opt/lolcat/bin/lolcat -p {}'.format(random.random()*10.0)
        else:
            lolcat = 'cat'
    else:
        lolcat = 'lolcat -p {}'.format(random.random()*10.0)

    os.system('{} | {} | {}'.format(fortune, cowsay, lolcat))


if __name__ == '__main__':
    whoop_de_doo()



# -*- coding: utf-8 -*-

import getpass
import os
import re
import sys
import time
import uuid

from macdaily.util.compat import subprocess
from macdaily.util.const.macro import BOOLEAN_STATES, PASS, USER, VERSION
from macdaily.util.const.term import reset
from macdaily.util.tools.deco import check, retry


def get_boolean(environ, default=False):
    boolean = os.getenv(environ)
    if boolean is None:
        return default
    return BOOLEAN_STATES.get(boolean.strip().lower(), default)


def get_logfile():
    logfile = os.getenv('MACDAILY_LOGFILE')
    if logfile is None:
        dirname = os.path.join(get_logdir(), 'misc', VERSION)
        os.makedirs(dirname, exist_ok=True)
        filename = os.path.join(dirname, '{}-{!s}.log'.format(time.strftime(r"%Y%m%d-%H%M%S"), uuid.uuid4()))
        return filename
    return logfile


def get_logdir():
    return get_path('MACDAILY_LOGDIR', '~/Library/Logs/MacDaily')


@retry('N')
def get_input(confirm, prompt='Input: ', *, prefix='', suffix='', queue=None):
    if sys.stdin.isatty():
        try:
            return input('{}{}'.format(prompt, suffix))
        except KeyboardInterrupt:
            print(reset)
            raise
    try:
        subprocess.check_call(['osascript', confirm, '{}{}'.format(prefix, prompt)],
                              stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        RETURN = 'N'
    else:
        RETURN = 'Y'
    finally:
        if queue is not None:
            queue.put(RETURN)
    return RETURN


def get_int(environ, default=0):
    integer = os.getenv(environ)
    if integer is None:
        return default
    integer = re.sub(r'[,_]', '', integer.strip(), flags=re.IGNORECASE)
    try:
        if re.match(r'0x[0-9a-f]+', integer, re.IGNORECASE):
            return int(integer, base=16)
        if re.match(r'0o[0-7]+', integer, re.IGNORECASE):
            return int(integer, base=8)
        if re.match(r'0[0-7]+', integer, re.IGNORECASE):
            return int('0o{}'.format("".join(integer[2:])), base=8)
        if re.match(r'0b[01]+', integer, re.IGNORECASE):
            return int(integer, base=2)
        return int(integer)
    except ValueError:
        return default


@check
@retry(PASS)
def get_pass(askpass, queue=None):
    if sys.stdin.isatty():
        try:
            return getpass.getpass(prompt='Password:')
        except KeyboardInterrupt:
            print(reset)
            raise
    RETURN = subprocess.check_output([askpass, '🔑 Enter your password for {}.'.format(USER)],  # pylint: disable=E1101
                                     stderr=subprocess.DEVNULL).strip().decode()
    if queue is not None:
        queue.put(RETURN)
    return RETURN


def get_path(environ, default='.'):
    path = os.getenv(environ)
    if path is None:
        return os.path.realpath(os.path.expanduser(default))
    return os.path.realpath(os.path.expanduser(path.strip()))



# -*- coding: utf-8 -*-

import os
import shutil
import sys
import tempfile
import traceback

from macdaily.cmd.logging import LoggingCommand
from macdaily.core.cask import CaskCommand
from macdaily.util.compat import subprocess
from macdaily.util.const.term import bold, flash, purple_bg, red, red_bg, reset, under
from macdaily.util.tools.get import get_logfile
from macdaily.util.tools.make import make_stderr
from macdaily.util.tools.print import print_info, print_scpt, print_term, print_text
from macdaily.util.tools.script import script


class CaskLogging(CaskCommand, LoggingCommand):

    @property
    def log(self):
        return 'Brewfile'

    @property
    def ext(self):
        return ''

    def _check_exec(self):
        try:
            subprocess.check_call(['brew', 'command', 'bundle'],
                                  stdout=subprocess.DEVNULL, stderr=make_stderr(self._vflag))
        except subprocess.CalledProcessError:
            print_text(traceback.format_exc(), self._file, redirect=self._vflag)
            print('macdaily-{}: {}{}cask{}: command not found'.format(self.cmd, red_bg, flash, reset), file=sys.stderr)
            text = ('macdaily-{}: {}cask{}: you may find Bundler on '
                    '{}{}https://github.com/Homebrew/homebrew-bundle{}, '
                    'or install Bundler through following command -- '
                    "`{}brew tap homebrew/bundle{}'".format(self.cmd, red, reset, purple_bg, under, reset, bold, reset))
            print_term(text, self._file, redirect=self._qflag)
            return False
        self._var__exec_path = shutil.which('brew')
        return True

    def _parse_args(self, namespace):
        self._quiet = namespace.get('quiet', False)  # pylint: disable=attribute-defined-outside-init
        self._verbose = namespace.get('verbose', False)  # pylint: disable=attribute-defined-outside-init

    def _proc_logging(self, path):
        text = 'Listing installed {}'.format(self.desc[1])
        print_info(text, self._file, redirect=self._qflag)

        suffix = path.replace('/', ':')
        with tempfile.NamedTemporaryFile() as _temp_file:
            logfile = os.path.join(self._logroot, '{}-{}{}'.format(self.log, suffix, self.ext))
            argv = [path, 'bundle', 'dump', '--force', '--file={}'.format(_temp_file.name)]

            print_scpt(argv, self._file, redirect=self._qflag)
            script(argv, self._file, shell=True,
                   timeout=self._timeout, redirect=self._vflag)

            with open(_temp_file.name, 'r') as file:
                context = file.read()
            print_text(context, get_logfile(), redirect=self._vflag)

        with open(logfile, 'w') as file:
            file.writelines(filter(lambda s: s.startswith('cask'), context.strip().splitlines(True)))  # pylint: disable=filter-builtin-not-iterating



# -*- coding: utf-8 -*-

import os
import re
import sys

from macdaily.cli.help import get_help_parser, parse_args
from macdaily.util.const.macro import (CMD_ARCHIVE, CMD_BUNDLE,  # pylint: disable=unused-import
                                       CMD_CLEANUP, CMD_CONFIG, CMD_DEPENDENCY, CMD_HELP,
                                       CMD_INSTALL, CMD_LAUNCH, CMD_LOGGING, CMD_POSTINSTALL,
                                       CMD_REINSTALL, CMD_UNINSTALL, CMD_UPDATE, COMMANDS, MAP_DICT,
                                       MAP_ALL, MAP_ARCHIVE, MAP_BUNDLE, MAP_CLEANUP, MAP_COMMANDS,
                                       MAP_CONFIG, MAP_DEPENDENCY, MAP_HELP, MAP_INSTALL,
                                       MAP_LAUNCH, MAP_LOGGING, MAP_POSTINSTALL, MAP_REINSTALL,
                                       MAP_UNINSTALL, MAP_UPDATE, ROOT)
from macdaily.util.error import CommandNotImplemented


def help_(argv=None):
    # parse args
    args = parse_args(argv)

    if args.command is None:
        pth = os.path.join(ROOT, 'man/macdaily.8')
        os.execlp('man', 'man', pth)

    command = args.command.strip().lower()
    if command in MAP_COMMANDS:
        print(COMMANDS, end='')
        return

    # split args, fetch cmd & sub
    temp = command.split('-', maxsplit=1)
    if len(temp) == 2:
        cmd, sub = temp
    else:
        cmd, sub = temp[0], None

    def _find_help(cmd, sub, man):
        pth = None
        if sub is None:
            pth = os.path.join(ROOT, 'man/macdaily-{}.8'.format(cmd))
        if sub in man:
            pth = os.path.join(ROOT, 'man/macdaily-{}-{}.8'.format(cmd, MAP_DICT[sub]))
        if pth is None:
            CMD = globals().get('CMD_{}'.format(cmd.upper()), set())
            parser = get_help_parser()
            pattern = r'.*{}.*'.format(command)
            matches = "', '".format().join(filter(lambda s: re.match(pattern, s, re.IGNORECASE),  # pylint: disable=cell-var-from-loop
                                          (r'%s-%s' % (cmd, sub) for sub in CMD)))
            if matches:
                parser.error("argument CMD: invalid choice: {!r} "
                             "(did you mean: '{}')".format(args.command, matches))
            else:
                parser.error("argument CMD: invalid choice: {!r} "
                             "(choose from {}-{})".format(args.command, cmd, (', %s-' % cmd).join(sorted(CMD))))
        os.execlp('man', 'man', pth)

    if cmd in MAP_ARCHIVE:
        _find_help('archive', sub, CMD_ARCHIVE)
    elif cmd in MAP_BUNDLE:
        # _find_help('bundle', sub, CMD_BUNDLE)
        raise CommandNotImplemented
    elif cmd in MAP_CLEANUP:
        _find_help('cleanup', sub, CMD_CLEANUP)
    elif cmd in MAP_CONFIG:
        _find_help('config', sub, CMD_CONFIG)
    elif cmd in MAP_DEPENDENCY:
        _find_help('dependency', sub, CMD_DEPENDENCY)
    elif cmd in MAP_HELP:
        _find_help('help', sub, CMD_HELP)
    elif cmd in MAP_INSTALL:
        _find_help('install', sub, CMD_INSTALL)
    elif cmd in MAP_LAUNCH:
        _find_help('launch', sub, CMD_LAUNCH)
    elif cmd in MAP_LOGGING:
        _find_help('logging', sub, CMD_LOGGING)
    elif cmd in MAP_POSTINSTALL:
        _find_help('postinstall', sub, CMD_POSTINSTALL)
    elif cmd in MAP_REINSTALL:
        _find_help('reinstall', sub, CMD_REINSTALL)
    elif cmd in MAP_UNINSTALL:
        _find_help('uninstall', sub, CMD_UNINSTALL)
    elif cmd in MAP_UPDATE:
        _find_help('update', sub, CMD_UPDATE)
    else:
        parser = get_help_parser()
        pattern = r'.*{}.*'.format(cmd)
        matches = "', '".join(filter(lambda s: re.match(pattern, s, re.IGNORECASE), MAP_ALL))  # pylint: disable=cell-var-from-loop
        if matches:
            parser.error('unrecognized arguments: {!r} '
                         "(did you mean: '{}')".format(args.command, matches))
        else:
            parser.error("argument CMD: invalid choice: {!r} "
                         "(choose from {})".format(args.command, ', '.join(sorted(MAP_ALL))))


if __name__ == '__main__':
    sys.exit(help_())



# -*- coding: utf-8 -*-

import argparse
import re
import sys

from macdaily.util.const.macro import (CMD_LOGGING, MAP_DICT, STR_APM, STR_APP, STR_BREW, STR_CASK,
                                       STR_GEM, STR_LOGGING, STR_MAS, STR_NPM, STR_PIP, STR_TAP)
from macdaily.util.const.macro import VERSION as __version__


def get_logging_parser():
    #######################################################
    # Logging CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    parser = argparse.ArgumentParser(prog='macdaily-logging',
                                     description='macOS Package Logging Automator',
                                     usage='macdaily logging [options] <mode-selection> ...',
                                     epilog=STR_LOGGING)
    parser.add_argument('-V', '--version', action='version', version=__version__)
    parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    genl_group = parser.add_argument_group(title='general arguments')
    genl_group.add_argument('-a', '--all', action='store_true',
                            help=('log all packages installed through Atom, RubyGems, Node.js, '
                                  'Homebrew, Caskroom, Mac App Store, and etc'))
    genl_group.add_argument('-q', '--quiet', action='store_true',
                            help='run in quiet mode, with no output information')
    genl_group.add_argument('-v', '--verbose', action='store_true',
                            help='run in verbose mode, with detailed output information')
    genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                            help='do not run cleanup process')
    genl_group.add_argument('-l', '--show-log', action='store_true',
                            help='open log in Console.app upon completion of command')

    ctrl_group = parser.add_argument_group(title='control arguments',
                                           description='options used to disable logging of certain mode')
    ctrl_group.add_argument('--no-apm', action='store_true', help='do not log Atom plug-ins')
    ctrl_group.add_argument('--no-app', action='store_true', help='do not log system applications')
    ctrl_group.add_argument('--no-gem', action='store_true', help='do not log Ruby gems')
    ctrl_group.add_argument('--no-mas', action='store_true', help='do not log macOS applications')
    ctrl_group.add_argument('--no-npm', action='store_true', help='do not log Node.js modules')
    ctrl_group.add_argument('--no-pip', action='store_true', help='do not log Python packages')
    ctrl_group.add_argument('--no-tap', action='store_true', help='do not log Homebrew Taps')
    ctrl_group.add_argument('--no-brew', action='store_true', help='do not log Homebrew formulae')
    ctrl_group.add_argument('--no-cask', action='store_true', help='do not log Homebrew Casks')

    parser.add_argument_group(title='mode selection',
                              description=('log existing packages installed through a specified method, '
                                           'e.g.: apm, app, gem, mas, npm, pip, tap, brew, cask'))

    return parser


def get_apm_parser():
    #######################################################
    # Apm Logging CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    apm_parser = argparse.ArgumentParser(prog='macdaily-logging-apm',
                                         description='Atom Plug-In Logging Automator',
                                         usage='macdaily logging apm [options] ...',
                                         epilog=STR_APM)
    apm_parser.add_argument('-V', '--version', action='version', version=__version__)
    apm_parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    apm_spec_group = apm_parser.add_argument_group(title='specification arguments')
    apm_spec_group.add_argument('-b', '--beta', action='store_true',
                                help='log Atom Beta plug-ins')

    apm_genl_group = apm_parser.add_argument_group(title='general arguments')
    apm_genl_group.add_argument('-q', '--quiet', action='store_true',
                                help='run in quiet mode, with no output information')
    apm_genl_group.add_argument('-v', '--verbose', action='store_true',
                                help='run in verbose mode, with detailed output information')
    apm_genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                                help='do not run cleanup process')
    apm_genl_group.add_argument('-l', '--show-log', action='store_true',
                                help='open log in Console.app upon completion of command')

    return apm_parser


def get_app_parser():
    #######################################################
    # App Logging CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    app_parser = argparse.ArgumentParser(prog='macdaily-logging-app',
                                         description='Mac Application Logging Automator',
                                         usage='macdaily logging app [options] ...',
                                         epilog=STR_APP)
    app_parser.add_argument('-V', '--version', action='version', version=__version__)
    app_parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    app_genl_group = app_parser.add_argument_group(title='general arguments')
    app_genl_group.add_argument('-q', '--quiet', action='store_true',
                                help='run in quiet mode, with no output information')
    app_genl_group.add_argument('-v', '--verbose', action='store_true',
                                help='run in verbose mode, with detailed output information')
    app_genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                                help='do not run cleanup process')
    app_genl_group.add_argument('-l', '--show-log', action='store_true',
                                help='open log in Console.app upon completion of command')

    return app_parser


def get_gem_parser():
    #######################################################
    # Gem Logging CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    gem_parser = argparse.ArgumentParser(prog='macdaily-logging-gem',
                                         description='Ruby Gem Logging Automator',
                                         usage='macdaily logging gem [options] ...',
                                         epilog=STR_GEM)
    gem_parser.add_argument('-V', '--version', action='version', version=__version__)
    gem_parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    gem_spec_group = gem_parser.add_argument_group(title='specification arguments')
    gem_spec_group.add_argument('-b', '--brew', action='store_true',
                                help='log gems of Ruby installed from Homebrew')
    gem_spec_group.add_argument('-s', '--system', action='store_true',
                                help='log gems of Ruby provided by macOS system')

    gem_genl_group = gem_parser.add_argument_group(title='general arguments')
    gem_genl_group.add_argument('-q', '--quiet', action='store_true',
                                help='run in quiet mode, with no output information')
    gem_genl_group.add_argument('-v', '--verbose', action='store_true',
                                help='run in verbose mode, with detailed output information')
    gem_genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                                help='do not run cleanup process')
    gem_genl_group.add_argument('-l', '--show-log', action='store_true',
                                help='open log in Console.app upon completion of command')

    return gem_parser


def get_mas_parser():
    #######################################################
    # Mas Logging CLI
    #   * options
    #       - optional arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    mas_parser = argparse.ArgumentParser(prog='macdaily-logging-mas',
                                         description='macOS Application Logging Automator',
                                         usage='macdaily logging mas [options] ...',
                                         epilog=STR_MAS)
    mas_parser.add_argument('-V', '--version', action='version', version=__version__)
    mas_parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    mas_genl_group = mas_parser.add_argument_group(title='general arguments')
    mas_genl_group.add_argument('-q', '--quiet', action='store_true',
                                help='run in quiet mode, with no output information')
    mas_genl_group.add_argument('-v', '--verbose', action='store_true',
                                help='run in verbose mode, with detailed output information')
    mas_genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                                help='do not run cleanup process')
    mas_genl_group.add_argument('-l', '--show-log', action='store_true',
                                help='open log in Console.app upon completion of command')

    return mas_parser


def get_npm_parser():
    #######################################################
    # Npm Logging CLI
    #   * options
    #       - optional arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    npm_parser = argparse.ArgumentParser(prog='macdaily-logging-npm',
                                         description='Node.js Module Logging Automator',
                                         usage='macdaily logging npm [options] ...',
                                         epilog=STR_NPM)
    npm_parser.add_argument('-V', '--version', action='version', version=__version__)
    npm_parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    npm_spec_group = npm_parser.add_argument_group(title='specification arguments')
    npm_spec_group.add_argument('-i', '--long', action='store_true',
                                help='show extended information')

    npm_genl_group = npm_parser.add_argument_group(title='general arguments')
    npm_genl_group.add_argument('-q', '--quiet', action='store_true',
                                help='run in quiet mode, with no output information')
    npm_genl_group.add_argument('-v', '--verbose', action='store_true',
                                help='run in verbose mode, with detailed output information')
    npm_genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                                help='do not run cleanup process')
    npm_genl_group.add_argument('-l', '--show-log', action='store_true',
                                help='open log in Console.app upon completion of command')

    return npm_parser


def get_pip_parser():
    #######################################################
    # Pip Logging CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    pip_parser = argparse.ArgumentParser(prog='macdaily-logging-pip',
                                         description='Python Package Logging Automator',
                                         usage='macdaily logging pip [options] ...',
                                         epilog=STR_PIP)
    pip_parser.add_argument('-V', '--version', action='version', version=__version__)
    pip_parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    pip_spec_group = pip_parser.add_argument_group(title='specification arguments')
    pip_spec_group.add_argument('-x', '--exclude-editable', action='store_true',
                                help='exclude editable package from output')
    pip_spec_group.add_argument('-b', '--brew', action='store_true',
                                help='log packages of Python installed from Homebrew')
    pip_spec_group.add_argument('-c', '--cpython', action='store_true',
                                help='log packages of CPython implementation')
    pip_spec_group.add_argument('-e', '--python', action='append', nargs='+',
                                default=list(), metavar='VER', dest='version',
                                help='indicate packages from which version of Python will be logged')
    pip_spec_group.add_argument('-r', '--pypy', action='store_true',
                                help='log packages of PyPy implementation')
    pip_spec_group.add_argument('-s', '--system', action='store_true',
                                help='log packages of Python provided by macOS system')

    pip_genl_group = pip_parser.add_argument_group(title='general arguments')
    pip_genl_group.add_argument('-q', '--quiet', action='store_true',
                                help='run in quiet mode, with no output information')
    pip_genl_group.add_argument('-v', '--verbose', action='store_true',
                                help='run in verbose mode, with detailed output information')
    pip_genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                                help='do not run cleanup process')
    pip_genl_group.add_argument('-l', '--show-log', action='store_true',
                                help='open log in Console.app upon completion of command')

    return pip_parser


def get_tap_parser():
    #######################################################
    # Tap Logging CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    tap_parser = argparse.ArgumentParser(prog='macdaily-logging-tap',
                                         description='Homebrew Tap Logging Automator',
                                         usage='macdaily logging tap [options] ...',
                                         epilog=STR_TAP)
    tap_parser.add_argument('-V', '--version', action='version', version=__version__)
    tap_parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    tap_genl_group = tap_parser.add_argument_group(title='general arguments')
    tap_genl_group.add_argument('-q', '--quiet', action='store_true',
                                help='run in quiet mode, with no output information')
    tap_genl_group.add_argument('-v', '--verbose', action='store_true',
                                help='run in verbose mode, with detailed output information')
    tap_genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                                help='do not run cleanup process')
    tap_genl_group.add_argument('-l', '--show-log', action='store_true',
                                help='open log in Console.app upon completion of command')

    return tap_parser


def get_brew_parser():
    #######################################################
    # Brew Logging CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    brew_parser = argparse.ArgumentParser(prog='macdaily-logging-brew',
                                          description='Homebrew Formula Logging Automator',
                                          usage='macdaily logging brew [options] ...',
                                          epilog=STR_BREW)
    brew_parser.add_argument('-V', '--version', action='version', version=__version__)
    brew_parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    brew_genl_group = brew_parser.add_argument_group(title='general arguments')
    brew_genl_group.add_argument('-q', '--quiet', action='store_true',
                                 help='run in quiet mode, with no output information')
    brew_genl_group.add_argument('-v', '--verbose', action='store_true',
                                 help='run in verbose mode, with detailed output information')
    brew_genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                                 help='do not run cleanup process')
    brew_genl_group.add_argument('-l', '--show-log', action='store_true',
                                 help='open log in Console.app upon completion of command')

    return brew_parser


def get_cask_parser():
    #######################################################
    # Cask Logging CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    cask_parser = argparse.ArgumentParser(prog='macdaily-logging-cask',
                                          description='Homebrew Cask Logging Automator',
                                          usage='macdaily logging cask [options] ...',
                                          epilog=STR_CASK)
    cask_parser.add_argument('-V', '--version', action='version', version=__version__)
    cask_parser.add_argument('more_opts', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    cask_genl_group = cask_parser.add_argument_group(title='general arguments')
    cask_genl_group.add_argument('-q', '--quiet', action='store_true',
                                 help='run in quiet mode, with no output information')
    cask_genl_group.add_argument('-v', '--verbose', action='store_true',
                                 help='run in verbose mode, with detailed output information')
    cask_genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                                 help='do not run cleanup process')
    cask_genl_group.add_argument('-l', '--show-log', action='store_true',
                                 help='open log in Console.app upon completion of command')

    return cask_parser


def parse_args(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    # main parser process
    main_parser = get_logging_parser()
    main_args = main_parser.parse_args(argv or ['--help'])

    def _update(d, e):
        for k, v in e.items():
            if k in d:
                if isinstance(v, list):
                    d[k].extend(v)
                else:
                    d[k] = v
            else:
                d[k] = v
        return d

    # recursively parse mode options
    more_opts = main_args.more_opts
    while more_opts:
        # traverse all extra options
        for index, option in enumerate(more_opts, start=1):
            # ignore ``--`` (end of option list)
            if option == '--':
                continue

            # check if legal mode
            parser_name = MAP_DICT.get(option.lower(), 'null')
            get_parser = globals().get('get_{}_parser'.format(parser_name))
            if get_parser is None:
                pattern = r'.*{}.*'.format(option)
                matches = "', '".format().join(filter(lambda s: re.match(pattern, s, re.IGNORECASE), CMD_LOGGING))  # pylint: disable=cell-var-from-loop
                if matches:
                    main_parser.error('unrecognized arguments: {!r} '
                                      "(did you mean: '{}')".format(option, matches))
                else:
                    main_parser.error('unrecognized arguments: {!r}'.format(option))

            # parse mode arguments
            parser = get_parser()
            args = parser.parse_args(more_opts[index:] or ['--help'])

            # store/update parsed arguments
            opt_dict = getattr(main_args, option, dict())
            opt_dict.pop('more_opts', None)
            setattr(main_args, option, _update(opt_dict, vars(args)))

            # check for extra options
            more_opts = args.more_opts
            break

    delattr(main_args, 'more_opts')
    return main_args



# -*- coding: utf-8 -*-

import contextlib
import functools
import pathlib
import sys


@functools.lru_cache(maxsize=128)
def find(root):
    with contextlib.suppress(PermissionError):
        for path in root.iterdir():
            with contextlib.suppress(OSError):
                if path.is_symlink():
                    continue
                if path.is_dir():
                    if path.parts[:2] == ('/', 'Volumes'):
                        continue
                    if path.suffix == '.app':
                        print(path)
                    find(path)


if __name__ == '__main__':
    sys.tracebacklimit = 0
    sys.exit(find(pathlib.Path(sys.argv[1]).resolve(strict=True)))



# -*- coding: utf-8 -*-

import abc
import collections
import contextlib
import copy
import glob
import os
import re
import sys
import traceback

from macdaily.cls.command import Command
from macdaily.util.compat import subprocess
from macdaily.util.const.term import bold, green, red, reset, yellow
from macdaily.util.tools.get import get_input
from macdaily.util.tools.make import make_stderr
from macdaily.util.tools.misc import date
from macdaily.util.tools.print import print_info, print_scpt, print_term, print_text
from macdaily.util.tools.script import sudo


class PipCommand(Command):

    @property
    def mode(self):
        return 'pip'

    @property
    def name(self):
        return 'Pip Installs Packages'

    @property
    def desc(self):
        return ('Python package', 'Python packages')

    def _check_exec(self):
        return True

    def _pkg_args(self, namespace):
        def match(version):
            return re.fullmatch(r'\d(\.\d)?', version)

        temp_ver = list()
        args_ver = namespace.get('version', list())
        for item in args_ver:
            if isinstance(item, str):
                item = filter(match, item.split(','))  # pylint: disable=filter-builtin-not-iterating
            for version in map(lambda s: s.split(','), item):
                temp_ver.extend(filter(match, version))
        self._version = set(temp_ver)

        return super()._pkg_args(namespace)

    @abc.abstractmethod
    def _parse_args(self, namespace):
        super()._parse_args(namespace)
        self._brew = namespace.get('brew', False)  # pylint: disable=attribute-defined-outside-init
        self._cpython = namespace.get('cpython', False)  # pylint: disable=attribute-defined-outside-init
        self._no_cleanup = namespace.get('no_cleanup', False)  # pylint: disable=attribute-defined-outside-init
        self._pypy = namespace.get('pypy', False)  # pylint: disable=attribute-defined-outside-init
        self._system = namespace.get('system', False)  # pylint: disable=attribute-defined-outside-init
        self._verbose = namespace.get('verbose', False)  # pylint: disable=attribute-defined-outside-init

    def _loc_exec(self):
        EXEC_PATH = dict(
            source=dict(
                brew=collections.defaultdict(list),
                system=collections.defaultdict(list),
            ),
            version=collections.defaultdict(list),
            combination={
                (True, True): collections.defaultdict(list),
                (True, False): collections.defaultdict(list),
                (False, True): collections.defaultdict(list),
                (False, False): collections.defaultdict(list),
            },
            implementation=dict(
                cpython=collections.defaultdict(list),
                pypy=collections.defaultdict(list),
            ),
        )

        def _sort_glob(path, *, flag):
            implementation = 'cpython' if flag else 'pypy'
            glob_path = glob.glob(path)
            if not glob_path:
                return
            glob_path.sort(reverse=True)
            path = glob_path[0]

            try:
                ver_proc = subprocess.check_output([path, '--version'], stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError:
                return
            if re.match(rb"^Python \d\.\d", ver_proc, flags=(re.I | re.M)) is None:
                return
            ver_text = ver_proc.decode().split()[1]

            def _append_path(exec_path):
                exec_path[ver_text[:3]].append(path)
                exec_path[ver_text[0]].append(path)
                exec_path['main'].append(path)

            _append_path(EXEC_PATH['implementation'][implementation])
            _append_path(EXEC_PATH['combination'][(True, flag)])
            _append_path(EXEC_PATH['source']['brew'])
            _append_path(EXEC_PATH['version'])

        try:
            proc = subprocess.check_output(['brew', '--prefix'], stderr=make_stderr(self._vflag))
        except subprocess.CalledProcessError:
            prefix = '/usr/local'
        else:
            prefix = proc.decode().strip()

        _sort_glob(os.path.join(prefix, 'Cellar/pypy/*/bin/pypy'), flag=False)
        _sort_glob(os.path.join(prefix, 'Cellar/pypy3/*/bin/pypy3'), flag=False)
        _sort_glob(os.path.join(prefix, 'Cellar/python@2/*/bin/python2.?'), flag=True)
        _sort_glob(os.path.join(prefix, 'Cellar/python/*/bin/python3.?'), flag=True)

        def _append_path(exec_path):
            exec_path[version[0]].append(path)  # pylint: disable=undefined-loop-variable
            exec_path[version].append(path)  # pylint: disable=undefined-loop-variable
            exec_path['main'].append(path)  # pylint: disable=undefined-loop-variable

        for path in glob.glob('/Library/Frameworks/Python.framework/Versions/?.?/bin/python?.?'):
            version = path[-3:]
            _append_path(EXEC_PATH['combination'][(False, True)])
            _append_path(EXEC_PATH['implementation']['cpython'])
            _append_path(EXEC_PATH['source']['system'])
            _append_path(EXEC_PATH['version'])

        def _extend_version(exec_path):
            if self._version:
                for version in self._version:
                    temp_exec.extend(exec_path[version])
            else:
                temp_exec.extend(exec_path['main'])

        temp_exec = list()
        if not any([self._brew, self._cpython, self._pypy, self._system]):
            _extend_version(EXEC_PATH['version'])
        elif (any([self._brew, self._system]) and not any([self._cpython, self._pypy])):
            if self._brew:
                _extend_version(EXEC_PATH['source']['brew'])
            if self._system:
                _extend_version(EXEC_PATH['source']['system'])
        elif (any([self._cpython, self._pypy]) and not any([self._brew, self._system])):
            if self._cpython:
                _extend_version(EXEC_PATH['implementation']['cpython'])
            if self._pypy:
                _extend_version(EXEC_PATH['implementation']['pypy'])
        else:
            if self._brew and self._cpython:
                _extend_version(EXEC_PATH['combination'][(True, True)])
            if self._brew and self._pypy:
                _extend_version(EXEC_PATH['combination'][(True, False)])
            if self._system and self._cpython:
                _extend_version(EXEC_PATH['combination'][(False, True)])
            if self._system and self._pypy:
                _extend_version(EXEC_PATH['combination'][(False, False)])
        self._exec = set(temp_exec)

    def _check_pkgs(self, path):
        _temp_pkgs = list()
        _lost_pkgs = list()

        text = 'Checking existence of specified packages'
        print_info(text, self._file, redirect=self._vflag)
        for package in self._packages:
            argv = [path, '-m', 'pip', 'show', package]

            args = ' '.join(argv)
            print_scpt(args, self._file, redirect=self._vflag)
            with open(self._file, 'a') as file:
                file.write('Script started on {}\n'.format(date()))
                file.write('command: {!r}\n'.format(args))

            try:
                proc = subprocess.check_output(argv, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError:
                print_text(traceback.format_exc(), self._file, redirect=self._vflag)
                _lost_pkgs.append(package)
            else:
                print_text(proc.decode(), self._file, redirect=self._vflag)
                _temp_pkgs.append(package)
            finally:
                with open(self._file, 'a') as file:
                    file.write('Script done on {}\n'.format(date()))

        if _lost_pkgs:
            text = 'Listing installed {}'.format(self.desc[1])
            print_info(text, self._file, redirect=self._vflag)

            self._lost.extend(_lost_pkgs)
            argv = [path, '-m', 'pip', 'list']

            args = ' '.join(argv)
            print_scpt(args, self._file, redirect=self._vflag)
            with open(self._file, 'a') as file:
                file.write('Script started on {}\n'.format(date()))
                file.write('command: {!r}\n'.format(args))

            try:
                proc = subprocess.check_output(argv, stderr=make_stderr(self._vflag))
            except subprocess.CalledProcessError:
                print_text(traceback.format_exc(), self._file, redirect=self._vflag)
                self._var__real_pkgs = set()  # pylint: disable=attribute-defined-outside-init
            else:
                context = proc.decode()
                print_text(context, self._file, redirect=self._vflag)
                self._var__real_pkgs = set(map(lambda pkg: pkg.split('==')[0], filter(None, context.splitlines())))  # pylint: disable=attribute-defined-outside-init,filter-builtin-not-iterating
            finally:
                with open(self._file, 'a') as file:
                    file.write('Script done on {}\n'.format(date()))
        else:
            self._var__real_pkgs = set()  # pylint: disable=attribute-defined-outside-init

        self._var__lost_pkgs = set(_lost_pkgs)  # pylint: disable=attribute-defined-outside-init
        self._var__temp_pkgs = set(_temp_pkgs)  # pylint: disable=attribute-defined-outside-init

    def _proc_fixmissing(self, path):
        text = 'Checking broken {} dependencies'.format(self.desc[0])
        print_info(text, self._file, redirect=self._qflag)

        def _proc_check():
            argv = [path, '-m', 'pip', 'check']
            args = ' '.join(argv)
            print_scpt(args, self._file, redirect=self._vflag)
            with open(self._file, 'a') as file:
                file.write('Script started on {}\n'.format(date()))
                file.write('command: {!r}\n'.format(args))

            _deps_pkgs = list()
            try:  # pip check exits with a non-zero status if any packages are missing dependencies
                proc = subprocess.run(argv, stdout=subprocess.PIPE, stderr=make_stderr(self._vflag))
            except subprocess.SubprocessError:
                print_text(traceback.format_exc(), self._file, redirect=self._vflag)
            else:
                context = proc.stdout.decode()
                print_text(context, self._file, redirect=self._vflag)

                for line in filter(None, context.strip().splitlines()):
                    if line == 'No broken requirements found.':
                        return set()
                    if 'which is not installed' in line:
                        _deps_pkgs.append(line.split()[3][:-1])
                    else:
                        _deps_pkgs.append(line.split()[4][:-1])
            finally:
                with open(self._file, 'a') as file:
                    file.write('Script done on {}\n'.format(date()))
            return set(_deps_pkgs)

        def _proc_confirm():
            pkgs = '{}, {}'.format(reset, bold).join(_deps_pkgs)
            text = 'macdaily-{}: {}pip{}: found broken dependencies: {}{}{}'.format(self.cmd, yellow, reset, bold, pkgs, reset)
            print_term(text, self._file, redirect=self._qflag)
            if self._yes or self._quiet:
                return True
            while True:
                ans = get_input(self._confirm, 'Would you like to reinstall?',
                                prefix='Found broken dependencies: {}.\n\n'.format(", ".join(_deps_pkgs)),
                                suffix=' ({}y{}/{}N{}) '.format(green, reset, red, reset))
                if re.match(r'[yY]', ans):
                    return True
                if re.match(r'[nN]', ans):
                    return False
                print('Invalid input.', file=sys.stderr)

        _deps_pkgs = _proc_check() - self._ignore
        if not _deps_pkgs:
            text = 'macdaily-{}: {}pip{}: no broken dependencies'.format(self.cmd, green, reset)
            print_term(text, self._file, redirect=self._qflag)
            return

        text = 'Fixing broken {} dependencies'.format(self.desc[0])
        print_info(text, self._file, redirect=self._qflag)

        if _proc_confirm():
            argv = [path, '-m', 'pip', 'reinstall']
            if self._quiet:
                argv.append('--quiet')
            if self._verbose:
                argv.append('--verbose')
            args = ' '.join(argv)

            _done_pkgs = set()
            while _deps_pkgs:
                for package in _deps_pkgs:
                    real_name = re.split(r'[<>=!]', package, maxsplit=1)[0]
                    print_scpt('{} {}'.format(args, package), self._file, redirect=self._qflag)

                    temp = copy.copy(argv)
                    temp[3] = 'uninstall'
                    print_scpt('{} {}'.format(" ".join(temp), real_name), self._file, redirect=self._qflag)
                    sudo('{} -m pip uninstall {} --yes'.format(path, real_name), self._file, self._password,
                         redirect=self._qflag, verbose=self._vflag, sethome=True, timeout=self._timeout)

                    temp = copy.copy(argv)
                    temp[3] = 'install'
                    print_scpt('{} {}'.format(" ".join(temp), package), self._file, redirect=self._qflag)
                    if not sudo('{} -m pip install {}'.format(path, package), self._file, self._password,
                                redirect=self._qflag, verbose=self._vflag, sethome=True, timeout=self._timeout):
                        with contextlib.suppress(ValueError):
                            self._pkgs.remove(real_name)
                _done_pkgs |= _deps_pkgs
                _deps_pkgs = _proc_check() - _done_pkgs - self._ignore
            text = 'macdaily-{}: {}pip{}: all broken dependencies fixed'.format(self.cmd, green, reset)
        else:
            text = 'macdaily-{}: {}pip{}: all broken dependencies remain'.format(self.cmd, red, reset)
        print_term(text, self._file, redirect=self._qflag)

    def _proc_cleanup(self):
        if self._no_cleanup:
            return

        text = 'Pruning caches and archives'
        print_info(text, self._file, redirect=self._qflag)

        argv = ['pip', 'cleanup']
        if self._verbose:
            argv.append('--verbose')
        if self._quiet:
            argv.append('--quiet')
        print_scpt(' '.join(argv), self._file, redirect=self._qflag)

        argv = ['rm', '-rf']
        if self._verbose:
            argv.append('-v')
        argv.append('')
        for path in ['/var/root/Library/Caches/pip/http/',
                     '/var/root/Library/Caches/pip/wheels/',
                     os.path.expanduser('~/Library/Caches/pip/http/'),
                     os.path.expanduser('~/Library/Caches/pip/wheels/')]:
            argv[-1] = path
            print_scpt(' '.join(argv), self._file, redirect=self._qflag)
            sudo(argv, self._file, self._password,
                 redirect=self._qflag, verbose=self._vflag)



# -*- coding: utf-8 -*-

import traceback

from macdaily.core.brew import BrewCommand
from macdaily.util.compat import subprocess
from macdaily.util.const.string import MAX, MIN
from macdaily.util.tools.misc import date
from macdaily.util.tools.print import print_info, print_scpt, print_text
from macdaily.util.tools.script import run


class PostinstallCommand(BrewCommand):

    @property
    def cmd(self):
        return 'postinstall'

    @property
    def act(self):
        return ('postinstall', 'postinstalled', 'postinstalled')

    @property
    def job(self):
        return ('postinstallation', 'postinstallation')

    def _parse_args(self, namespace):
        self._endswith = namespace.get('endswith', MAX)  # pylint: disable=attribute-defined-outside-init
        self._no_cleanup = namespace.get('no_cleanup', False)  # pylint: disable=attribute-defined-outside-init
        self._startswith = namespace.get('startswith', MIN)  # pylint: disable=attribute-defined-outside-init

        self._all = namespace.get('all', False)  # pylint: disable=attribute-defined-outside-init
        self._quiet = namespace.get('quiet', False)  # pylint: disable=attribute-defined-outside-init
        self._verbose = namespace.get('verbose', False)  # pylint: disable=attribute-defined-outside-init
        self._yes = namespace.get('yes', False)  # pylint: disable=attribute-defined-outside-init

        self._logging_opts = namespace.get('logging', str()).split()  # pylint: disable=attribute-defined-outside-init
        self._postinstall_opts = namespace.get('postinstall', str()).split()  # pylint: disable=attribute-defined-outside-init

    def _run_proc(self):
        self._pkgs = list()
        self._fail = list()
        self._lost = list()
        for path in self._exec:
            text = 'Using {} executable {!r}'.format(self.name, path)
            print_info(text, self._file, redirect=self._qflag)
            if self._proc_logging(path):
                self._proc_postinstall(path)
            else:
                text = 'No {} to postinstall for executable {!r}'.format(self.desc[1], path)
                print_info(text, self._file, redirect=self._qflag)
            self._proc_fixmissing(path)
        self._proc_cleanup()

    def _proc_logging(self, path):
        if self._packages:
            self._check_pkgs(path)
            self._did_you_mean()
        else:
            self._check_list(path)
        return self._check_confirm(path)

    def _check_list(self, path):
        text = 'Checking installed {}'.format(self.desc[1])
        print_info(text, self._file, redirect=self._vflag)

        argv = [path, 'list']
        argv.extend(self._logging_opts)

        args = ' '.join(argv)
        print_scpt(args, self._file, redirect=self._vflag)
        with open(self._file, 'a') as file:
            file.write('Script started on {}\n'.format(date()))
            file.write('command: {!r}\n'.format(args))

        try:
            proc = subprocess.check_output(argv, stderr=subprocess.DEVNULL)
        except subprocess.SubprocessError:
            print_text(traceback.format_exc(), self._file, redirect=self._vflag)
            self._var__temp_pkgs = set()  # pylint: disable=attribute-defined-outside-init
        else:
            context = proc.decode()
            print_text(context, self._file, redirect=self._vflag)

            _temp_pkgs = list()
            for package in context.strip().split():
                if self._startswith <= package <= self._endswith:
                    _temp_pkgs.append(package)
            self._var__temp_pkgs = set(_temp_pkgs)  # pylint: disable=attribute-defined-outside-init
        finally:
            with open(self._file, 'a') as file:
                file.write('Script done on {}\n'.format(date()))

    def _proc_postinstall(self, path):
        text = 'Postinstalling specified {}'.format(self.desc[1])
        print_info(text, self._file, redirect=self._qflag)

        argv = [path, 'postinstall']
        if self._quiet:
            argv.append('--quiet')
        if self._verbose:
            argv.append('--verbose')
        argv.extend(self._postinstall_opts)

        argv.append('')
        for package in self._var__temp_pkgs:
            argv[-1] = package
            print_scpt(' '.join(argv), self._file, redirect=self._qflag)
            if run(argv, self._file, timeout=self._timeout,
                   redirect=self._qflag, verbose=self._vflag):
                self._fail.append(package)
            else:
                self._pkgs.append(package)
        del self._var__temp_pkgs



# -*- coding: utf-8 -*-

import copy
import os
import textwrap
import time
import traceback

from macdaily.cmd.update import UpdateCommand
from macdaily.core.cask import CaskCommand
from macdaily.util.compat import subprocess
from macdaily.util.const.term import length
from macdaily.util.tools.make import make_stderr
from macdaily.util.tools.misc import date
from macdaily.util.tools.print import print_info, print_scpt, print_text
from macdaily.util.tools.script import run


class CaskUpdate(CaskCommand, UpdateCommand):

    def _parse_args(self, namespace):
        self._exhaust = namespace.get('exhaust', False)  # pylint: disable=attribute-defined-outside-init
        self._force = namespace.get('force', False)  # pylint: disable=attribute-defined-outside-init
        self._greedy = namespace.get('greedy', False)  # pylint: disable=attribute-defined-outside-init
        self._merge = namespace.get('merge', False)  # pylint: disable=attribute-defined-outside-init
        self._no_cleanup = namespace.get('no_cleanup', False)  # pylint: disable=attribute-defined-outside-init

        self._all = namespace.get('all', False)  # pylint: disable=attribute-defined-outside-init
        self._quiet = namespace.get('quiet', False)  # pylint: disable=attribute-defined-outside-init
        self._verbose = namespace.get('verbose', False)  # pylint: disable=attribute-defined-outside-init
        self._yes = namespace.get('yes', False)  # pylint: disable=attribute-defined-outside-init

        self._logging_opts = namespace.get('logging', str()).split()  # pylint: disable=attribute-defined-outside-init
        self._update_opts = namespace.get('update', str()).split()  # pylint: disable=attribute-defined-outside-init

    def _check_list(self, path):
        if (self._brew_renew is None or time.time() - self._brew_renew >= 300):
            self._proc_renew(path)
            self._brew_renew = time.time()

        if self._exhaust:
            self._exhaust_check(path)
            return

        text = 'Checking outdated {}'.format(self.desc[1])
        print_info(text, self._file, redirect=self._vflag)

        argv = [path, 'cask', 'outdated']
        if self._quiet:
            argv.append('--quiet')
        if self._verbose:
            argv.append('--verbose')
        if self._greedy:
            argv.append('--greedy')
        argv.extend(self._logging_opts)

        args = ' '.join(argv)
        print_scpt(args, self._file, redirect=self._vflag)
        with open(self._file, 'a') as file:
            file.write('Script started on {}\n'.format(date()))
            file.write('command: {!r}\n'.format(args))

        try:
            proc = subprocess.check_output(argv, stderr=make_stderr(self._vflag))
        except subprocess.SubprocessError:
            print_text(traceback.format_exc(), self._file, redirect=self._vflag)
            self._var__temp_pkgs = set()  # pylint: disable=attribute-defined-outside-init
        else:
            context = proc.decode()
            print_text(context, self._file, redirect=self._vflag)

            _temp_pkgs = list()
            for line in filter(None, context.strip().splitlines()):
                _temp_pkgs.append(line.split(maxsplit=1)[0])
            self._var__temp_pkgs = set(_temp_pkgs)  # pylint: disable=attribute-defined-outside-init
        finally:
            with open(self._file, 'a') as file:
                file.write('Script done on {}\n'.format(date()))

    def _exhaust_check(self, path):
        text = 'Checking outdated {} exclusively'.format(self.desc[1])
        print_info(text, self._file, redirect=self._vflag)

        argv = ['brew', 'cask', 'outdated', '--exhaust']
        if self._quiet:
            argv.append('--quiet')
        if self._verbose:
            argv.append('--verbose')
        argv.extend(self._logging_opts)
        print_scpt(' '.join(argv), self._file, redirect=self._vflag)

        text = 'Listing installed {}'.format(self.desc[1])
        print_info(text, self._file, redirect=self._vflag)

        argv = [path, 'cask', 'list']
        args = ' '.join(argv)
        print_scpt(args, self._file, redirect=self._vflag)
        with open(self._file, 'a') as file:
            file.write('Script started on {}\n'.format(date()))
            file.write('command: {!r}\n'.format(args))

        try:
            proc = subprocess.check_output(argv, stderr=make_stderr(self._vflag))
        except subprocess.CalledProcessError:
            print_text(traceback.format_exc(), self._file, redirect=self._vflag)
            _list_pkgs = set()
        else:
            context = proc.decode()
            _list_pkgs = set(context.split())
            print_text(context, self._file, redirect=self._vflag)
        finally:
            with open(self._file, 'a') as file:
                file.write('Script done on {}\n'.format(date()))

        text = 'Fetching Homebrew prefix'.format()
        print_info(text, self._file, redirect=self._vflag)

        argv = [path, '--prefix']
        args = ' '.join(argv)
        print_scpt(args, self._file, redirect=self._vflag)
        with open(self._file, 'a') as file:
            file.write('Script started on {}\n'.format(date()))
            file.write('command: {!r}\n'.format(args))

        fail = False
        try:
            proc = subprocess.check_output(argv, stderr=make_stderr(self._vflag))
        except subprocess.CalledProcessError:
            print_text(traceback.format_exc(), self._file, redirect=self._vflag)
            self._var__temp_pkgs = set()  # pylint: disable=attribute-defined-outside-init
            fail = True
        else:
            context = proc.decode()
            print_text(context, self._file, redirect=self._vflag)
            prefix = context.strip()
        finally:
            with open(self._file, 'a') as file:
                file.write('Script done on {}\n'.format(date()))
        if fail:
            return

        text = 'Checking versions of installed {}'.format(self.desc[1])
        print_info(text, self._file, redirect=self._vflag)

        _temp_pkgs = list()
        for cask in _list_pkgs:
            argv = [path, 'cask', 'info', cask]
            args = ' '.join(argv)
            print_scpt(args, self._file, redirect=self._vflag)
            with open(self._file, 'a') as file:
                file.write('Script started on {}\n'.format(date()))
                file.write('command: {!r}\n'.format(args))

            try:
                proc = subprocess.check_output(argv, stderr=make_stderr(self._vflag))
            except subprocess.CalledProcessError:
                print_text(traceback.format_exc(), self._file, redirect=self._vflag)
            else:
                context = proc.decode()
                print_text(context, self._file, redirect=self._vflag)

                version = context.split(maxsplit=2)[1]
                installed = os.path.join(prefix, 'Caskroom', cask, version)
                if os.path.isdir(installed):
                    _temp_pkgs.append(cask)
            finally:
                with open(self._file, 'a') as file:
                    file.write('Script done on {}\n'.format(date()))

        self._var__temp_pkgs = set(_temp_pkgs)  # pylint: disable=attribute-defined-outside-init
        max_len = len(max(_temp_pkgs, key=len))
        context = os.linesep.join(textwrap.wrap('    '.join(
            map(lambda s: s.ljust(max_len), self._var__temp_pkgs)), length))
        print_scpt('{} cask outdated list'.format(path), self._file, redirect=self._vflag)
        print_text(context, self._file, redirect=self._vflag)

    def _proc_update(self, path):
        text = 'Upgrading outdated {}'.format(self.desc[1])
        print_info(text, self._file, redirect=self._qflag)

        argv = [path, 'cask', 'upgrade']
        if self._quiet:
            argv.append('--quiet')
        if self._verbose:
            argv.append('--verbose')
        if self._greedy:
            argv.append('--greedy')
        argv.extend(self._update_opts)

        temp = copy.copy(argv)
        if self._exhaust:
            temp.append('--exhaust')
        args = ' '.join(temp)
        argv.append('')

        askpass = 'SUDO_ASKPASS={!r}'.format(self._askpass)
        for package in self._var__temp_pkgs:
            argv[-1] = package
            print_scpt('{} {}'.format(args, package), self._file, redirect=self._qflag)
            if run(argv, self._file, shell=True, timeout=self._timeout,
                   redirect=self._qflag, verbose=self._vflag, prefix=askpass):
                self._fail.append(package)
            else:
                self._pkgs.append(package)
        del self._var__temp_pkgs



# -*- coding: utf-8 -*-

import importlib
import os
import platform
import warnings
import sys

import tbtrim

from macdaily.util.const.macro import DEVMODE
from macdaily.util.const.term import red, reset
from macdaily.util.error import Error, UnsupportedOS
from macdaily.util.tools.misc import predicate

# set up sys.excepthook
if DEVMODE:
    warnings.showwarning('development mode enabled'.format(), RuntimeWarning,
                         filename=__file__, lineno=0,
                         line="MACDAILY_DEVMODE={}".format(os.environ['MACDAILY_DEVMODE']))
else:
    tbtrim.set_trim_rule(predicate, strict=False,
                         target=(Error, KeyboardInterrupt))

# check platform
if platform.system() != 'Darwin':
    raise UnsupportedOS('macdaily: error: script runs only on macOS')

# check dependencies
if sys.version_info[:2] <= (3, 4):
    def test_import(module):
        try:
            importlib.import_module(module)
        except ImportError:
            print('macdaily: {}error{}: broken dependency'.format(red, reset), file=sys.stderr)
            raise
    test_import('pathlib2')
    test_import('subprocess32')



# -*- coding: utf-8 -*-

import platform
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# check platform
if platform.system() != 'Darwin':
    sys.exit('macdaily: error: script runs only on macOS')

# README
# with open('./README.rst', encoding='utf-8') as file:
#     long_description = file.read()
long_description = """\
:Platform:
    OS X Yosemite, OS X El Capitan, macOS Sierra
    macOS High Sierra, macOS Mojave
:Language: Python (version ≥ 3.4)
:Environment: Console | Terminal

==========================================
MacDaily - macOS Automated Package Manager
==========================================

|download| |version| |status| |format|

|python| |implementation|

    Package day-care manager on macOS.

**MacDaily** is an all-in-one collection of console utility written in Python
with support of `PTY <https://en.wikipedia.org/wiki/Pseudo_terminal>`__.
Originally works as an automated housekeeper for Mac to update all packages
outdated, **MacDaily** is now fully functioned and end-user oriented.

**MacDaily** can manage packages of various distributions of
`Atom <https://atom.io>`__, `RubyGems <https://rubygems.org>`__,
`Homebrew <https://brew.sh>`__, `Python <https://pypy.org>`__,
`Node.js <https://nodejs.org>`__, and even macOS software updates (c.f.
``softwareupdate(8)``). Without being aware of everything about your Mac, one
can easily work around and manage packages out of no pain using **MacDaily**.

.. |download| image:: http://pepy.tech/badge/macdaily
   :target: http://pepy.tech/count/macdaily
.. |version| image:: https://img.shields.io/pypi/v/macdaily.svg
   :target: https://pypi.org/project/macdaily
.. |format| image:: https://img.shields.io/pypi/format/macdaily.svg
   :target: https://pypi.org/project/macdaily
.. |status| image:: https://img.shields.io/pypi/status/macdaily.svg
   :target: https://pypi.org/project/macdaily
.. |python| image:: https://img.shields.io/pypi/pyversions/macdaily.svg
   :target: https://python.org
.. |implementation| image:: https://img.shields.io/pypi/implementation/macdaily.svg
   :target: http://pypy.org
"""

# version string
__version__ = '2019.4.7.post1'
# context = pkg_resources.resource_string(__name__, 'macdaily/util/const/macro.py')
# for line in context.splitlines():
#     match = re.match(rb"VERSION = '(.*)'", line)
#     if match is None:
#         continue
#     __version__ = match.groups()[0].decode()
#     break

# set-up script for pip distribution
setup(
    name='macdaily',
    version=__version__,
    author='Jarry Shaw',
    author_email='jarryshaw@icloud.com',
    url='https://github.com/JarryShaw/MacDaily#macdaily',
    license='Apple Public Source License',
    keywords='daily utility script',
    description='macOS Automated Package Manager',
    long_description=long_description,
    # long_description=pkg_resources.resource_string(__name__, 'README.rst').decode(),
    long_description_content_type='text/x-rst; charset=UTF-8',
    python_requires='>=3.4',
    # include_package_data=True,
    # zip_safe=True,
    install_requires=[
        'tbtrim>=0.2.1',
    ],
    extras_require={
        ':python_version <= "3.4"': ['pathlib2>=2.3.2', 'subprocess32>=3.5.3'],
        'all': ['ptyng>=0.3.3', 'dictdumper>=0.7.0.post1', 'configupdater'],
        'ptyng': ['ptyng>=0.3.3'],
        'tree': ['dictdumper>=0.7.0.post1'],
        'config': ['configupdater'],
    },
    entry_points={
        'console_scripts': [
            'macdaily = macdaily.__main__:main',
            'md-archive = macdaily.api.archive:archive',
            # 'md-bundle = macdaily.api.bundle:bundle',
            'md-cleanup = macdaily.api.cleanup:cleanup',
            'md-config = macdaily.api.config:config [config]',
            'md-dependency = macdaily.api.dependency:dependency [tree]',
            'md-help = macdaily.api.help:help_',
            'md-install = macdaily.api.install:install',
            'md-launch = macdaily.api.launch:launch',
            'md-logging = macdaily.api.logging:logging',
            'md-postinstall = macdaily.api.postinstall:postinstall',
            'md-reinstall = macdaily.api.reinstall:reinstall',
            'md-uninstall = macdaily.api.uninstall:uninstall',
            'md-update = macdaily.api.update:update',
        ]
    },
    # packages=setuptools.find_namespace_packages(
    #     include=['macdaily', 'macdaily.*'],
    # ),
    packages=[
        'macdaily',
        'macdaily.api',
        'macdaily.cli',
        'macdaily.cls',
        'macdaily.cls.bundle',
        'macdaily.cls.cleanup',
        'macdaily.cls.dependency',
        'macdaily.cls.install',
        'macdaily.cls.logging',
        'macdaily.cls.reinstall',
        'macdaily.cls.uninstall',
        'macdaily.cls.update',
        'macdaily.cmd',
        'macdaily.comp',
        'macdaily.core',
        'macdaily.man',
        'macdaily.img',
        'macdaily.res',
        'macdaily.util',
        'macdaily.util.const',
        'macdaily.util.tools',
    ],
    package_data={
        '': [
            'LICENSE',
            'README.rst',
        ],
        'macdaily.comp': ['macdaily.bash-completion'],
        'macdaily.man': ['*.8'],
        'macdaily.img': ['*.icns'],
        'macdaily.res': ['*.py', '*.applescript'],
    },
    classifiers=[
        # 'Development Status :: 7 - Inactive',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 5 - Production/Stable',
        'Development Status :: 4 - Beta',
        # 'Development Status :: 3 - Alpha',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 1 - Planning',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apple Public Source License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        # 'Programming Language :: Unix Shell',
        'Topic :: Utilities',
    ]
)



# -*- coding: utf-8 -*-

import argparse
import sys

from macdaily.util.const.macro import STR_LAUNCH
from macdaily.util.const.macro import VERSION as __version__
from macdaily.util.const.term import bold, reset


def get_launch_parser():
    #######################################################
    # Launch CLI
    #   * options
    #       - optional arguments
    #       - specification arguments
    #       - general arguments
    #       - miscellaneous arguments
    #   * packages
    #######################################################

    parser = argparse.ArgumentParser(prog='macdaily-launch',
                                     description='MacDaily Dependency Launch Helper',
                                     usage='macdaily launch [options] <prog-selection> ...',
                                     epilog=STR_LAUNCH)
    parser.add_argument('-V', '--version', action='version', version=__version__)

    spec_group = parser.add_argument_group(title='specification arguments')
    spec_group.add_argument('program', nargs='*', metavar='PROG',
                            help=("helper program to launch, choose from `{}askpass{}', "
                                  "`{}confirm{}' and `{}daemons{}'".format(bold, reset, bold, reset, bold, reset)))

    genl_group = parser.add_argument_group(title='general arguments')
    genl_group.add_argument('-a', '--all', action='store_true',
                            help=("launch all help programs, i.e. `{}askpass{}', "
                                  "`{}confirm{}' and `{}daemons{}'".format(bold, reset, bold, reset, bold, reset)))
    genl_group.add_argument('-q', '--quiet', action='store_true',
                            help='run in quiet mode, with no output information')
    genl_group.add_argument('-v', '--verbose', action='store_true',
                            help='run in verbose mode, with detailed output information')
    genl_group.add_argument('-n', '--no-cleanup', action='store_true',
                            help='do not run cleanup process')
    genl_group.add_argument('-l', '--show-log', action='store_true',
                            help='open log in Console.app upon completion of command')

    return parser


def parse_args(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    # main parser process
    main_parser = get_launch_parser()
    main_args = main_parser.parse_args(argv or ['--help'])

    return main_args



# -*- coding: utf-8 -*-

import contextlib
import os
import re
import shlex
import sys

from macdaily.util.const.term import blue, bold, dim, grey, length, purple, reset, under


def print_environ(file=sys.stdout, value_only=False, no_term=False,
                  prefix='  - ', suffix=' {}({}value{}{}: %s){}'.format(grey, under, reset, grey, reset)):
    if no_term:
        bold = green = reset = ''  # pylint: disable=redefined-outer-name
    else:
        from macdaily.util.const.term import bold, green, reset  # pylint: disable=reimported

    with contextlib.redirect_stdout(file):
        if not value_only:
            print('The following environment variables can be set, to do various things:')
            print()

        def print_value(environ, default=None):
            env = os.getenv(environ, default)
            if env is None:
                value = 'null'
            else:
                value = shlex.quote(env)
            print('{}{}{}{}{}'.format(prefix, bold, environ, reset, suffix) % value)

        print_value('SUDO_PASSWORD')
        print_value('NULL_PASSWORD', 'false')
        print_value('MACDAILY_NO_CHECK', 'false')
        print_value('MACDAILY_NO_CONFIG', 'false')
        print_value('MACDAILY_LOGDIR', '~/Library/Logs/MacDaily')
        print_value('MACDAILY_DSKDIR')
        print_value('MACDAILY_ARCDIR', '${MACDAILY_DSKDIR}/Developers')
        print_value('MACDAILY_LIMIT', '1,000')
        print_value('MACDAILY_RETRY', '60')
        print_value('MACDAILY_CLEANUP', 'true')
        print_value('MACDAILY_APM', 'true')
        print_value('MACDAILY_APP', 'true')
        print_value('MACDAILY_BREW', 'true')
        print_value('MACDAILY_CASK', 'true')
        print_value('MACDAILY_GEM', 'true')
        print_value('MACDAILY_MAS', 'true')
        print_value('MACDAILY_NPM', 'true')
        print_value('MACDAILY_PIP', 'true')
        print_value('MACDAILY_SYSTEM', 'true')
        print_value('MACDAILY_TAP', 'true')
        print_value('MACDAILY_DEVMODE', 'false')

        if not value_only:
            print()
            print('You can learn more at:')
            print('  {}{}https://github.com/JarryShaw/MacDaily#environment{}'.format(green, under, reset))


def print_info(text, file, redirect=False):
    flag = text.endswith(os.linesep)
    if not redirect:
        end = str() if flag else os.linesep
        print_wrap('{}{}|💼|{} {}{}{}'.format(bold, blue, reset, bold, text, reset), end=end)
    with open(file, 'a') as fd:
        context = re.sub(r'(\033\[[0-9][0-9;]*m)|(\^D\x08\x08)', r'',
                         (text if flag else '{}{}'.format(text, os.linesep)), flags=re.IGNORECASE)
        fd.write('|💼| {}'.format(context))


def print_misc(text, file, redirect=False):
    flag = text.endswith(os.linesep)
    if not redirect:
        end = str() if flag else os.linesep
        print_wrap('{}{}|📌|{} {}{}{}'.format(bold, grey, reset, bold, text, reset), end=end)
    with open(file, 'a') as fd:
        context = re.sub(r'(\033\[[0-9][0-9;]*m)|(\^D\x08\x08)', r'',
                         (text if flag else '{}{}'.format(text, os.linesep)), flags=re.IGNORECASE)
        fd.write('|📌| {}'.format(context))


def print_scpt(text, file, redirect=False):
    if not isinstance(text, str):
        text = ' '.join(text)
    flag = text.endswith(os.linesep)
    if not redirect:
        end = str() if flag else os.linesep
        print_wrap('{}{}|📜|{} {}{}{}'.format(bold, purple, reset, bold, text, reset), end=end)
    with open(file, 'a') as fd:
        context = re.sub(r'(\033\[[0-9][0-9;]*m)|(\^D\x08\x08)', r'',
                         (text if flag else '{}{}'.format(text, os.linesep)), flags=re.IGNORECASE)
        fd.write('|📜| {}'.format(context))


def print_term(text, file, redirect=False):
    flag = text.endswith(os.linesep)
    if not redirect:
        end = str() if flag else os.linesep
        print_wrap(text, end=end)
    with open(file, 'a') as fd:
        context = re.sub(r'(\033\[[0-9][0-9;]*m)|(\^D\x08\x08)', r'',
                         (text if flag else '{}{}'.format(text, os.linesep)), flags=re.IGNORECASE)
        fd.write(context)


def print_text(text, file, redirect=False):
    flag = text.endswith(os.linesep)
    if not redirect:
        end = str() if flag else os.linesep
        print_wrap('{}{}{}'.format(dim, text, reset), end=end)
    with open(file, 'a') as fd:
        context = re.sub(r'(\033\[[0-9][0-9;]*m)|(\^D\x08\x08)', r'',
                         (text if flag else '{}{}'.format(text, os.linesep)), flags=re.IGNORECASE)
        fd.write(context)


def print_wrap(text, length=length, **kwargs):  # pylint: disable=redefined-outer-name
    print(wrap_text(text, length), **kwargs)


def wrap_text(string, length=length):  # pylint: disable=redefined-outer-name,unused-argument
    # text = '\n'.join(textwrap.wrap(string, length))
    # if string.endswith(os.linesep):
    #     return f'{text}{os.linesep}'
    # return text
    return string
